from .filedproperty import *
from .plotting import *
from .sourcedetection import *
from .sourcepairing import *
from .astrometry import *

import os
import psutil
import numpy as np
import scipy as sp
import scipy.stats 
import math
import warnings

import logging
logger = logging.getLogger(__name__)

import typing
from numpy.typing import NDArray
from typing import Sequence, Union, Tuple, Any
if typing.TYPE_CHECKING:
    from iop4lib.db import ReducedFit, AstroSource

def get_column_values(qs, column_names):
    """ Given a queryset and a list of column names, return a dictionary with the values of each column as a numpy array

    None values are converted to np.nan (to avoid numpy arrays having dtype=object) 
    """
    values_lists = zip(*qs.values_list(*column_names))
    values_lists = [[x if x is not None else np.nan for x in values] for values in values_lists]
    return {k: v for k, v in zip(column_names, map(np.array, values_lists))}


def qs_to_table(qs=None, model=None, data=None, column_names=None, default_column_names=None):
    """ Specify either queryset, or data and models."""

    from django.db import models

    if data is None or model is None:
        if qs is None:
            raise Exception("Either qs or data and model must be specified")
        else:
            model = qs.model
            data = qs.values()

    if column_names is None:
        column_names = [f.name for f in model._meta.get_fields() if hasattr(f, 'name') and f.name in data[0].keys()]

    if default_column_names is None:
        default_column_names = column_names

    columns = [{
                    "name": k, 
                    "title": model._meta.get_field(k).verbose_name, 
                    "visible": (k in default_column_names),
                    "type": "int" if isinstance(model._meta.get_field(k), models.IntegerField) else \
                            "float" if isinstance(model._meta.get_field(k), models.FloatField) else \
                            "str" if isinstance(model._meta.get_field(k), models.CharField) else \
                            "str" if isinstance(model._meta.get_field(k), models.TextField) else \
                            "date" if isinstance(model._meta.get_field(k), models.DateField) else \
                            "unknown",
                    "help": model._meta.get_field(k).help_text,
                } for k in column_names]
    
    return {'data': list(data), 'columns': columns}

def divisorGenerator(n):
    """Generator for divisors of n"""
    large_divisors = []
    for i in range(1, int(np.sqrt(n) + 1)):
        if n % i == 0:
            yield i
            if i*i != n:
                large_divisors.append(n / i)
    for divisor in reversed(large_divisors):
        yield divisor


def stats_dict(data):
    """Return a dictionary with some basic statistics of the input data"""

    if isinstance(data, np.ma.MaskedArray):
        data = data.compressed()
    else:
            data = data.flatten()

    res = dict()

    res['mean'] = np.mean(data)
    res['median'] = np.median(data)
    res['std'] = np.std(data)
    res['min'] = np.min(data)
    res['max'] = np.max(data)
    res['mode'] = sp.stats.mode(data, axis=None, keepdims=False).mode

    return res


# Functions to get memory usage

def get_mem_children():
    """Return the sum of the memory usage of all children processes, from the parent process (in bytes)"""
    children = psutil.Process(os.getpid()).children(recursive=True)
    memory_usage = 0
    for child in children:
        try:
            mem_info = child.memory_info()
            memory_usage += mem_info.rss
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass
    return memory_usage

def get_mem_current():
    """Return the memory usage of the current process (in bytes)"""
    process = psutil.Process(os.getpid())
    mem_info = process.memory_info()
    return mem_info.rss

def get_mem_parent_from_child():
    """Return the memory usage of the parent process, from one of the child processes (in bytes)"""
    process = psutil.Process(os.getpid()).parent()
    mem_info = process.memory_info()
    return mem_info.rss

def get_mem_children_from_child():
    """Return the sum of the memory usage of all children processes, from one of the child processes (in bytes)"""
    parent = psutil.Process(os.getpid()).parent()
    children = parent.children(recursive=True)
    memory_usage = 0
    for child in children:
        try:
            mem_info = child.memory_info()
            memory_usage += mem_info.rss
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass
    return memory_usage

def get_total_mem_from_child():
    """Return the sum of the memory usage of all processes (parent and children) from one of the child processes (in bytes)"""
    return get_mem_parent_from_child() + get_mem_children_from_child()









# Function to get target FWHM

def estimate_common_apertures(redfL, reductionmethod=None):
    r"""estimate an appropriate common aperture for a list of reduced fits.
    
    It fits the target source profile in the fields and returns some multiples of the fwhm which are used as the aperture and as the inner and outer radius of the annulus for local bkg estimation).
    """

    import numpy as np
    from iop4lib.db import AstroSource
    from photutils.profiles import RadialProfile
    from photutils.centroids import centroid_quadratic
    from astropy.modeling.models import Moffat1D, Const1D, Gaussian1D
    from astropy.modeling import models, fitting

    astrosource_S = set.union(*[set(redf.sources_in_field.all()) for redf in redfL])
    target_L = [astrosource for astrosource in astrosource_S if not astrosource.is_calibrator]

    logger.debug(f"{astrosource_S=}")

    if len(target_L) > 0:
        target = target_L[0]
    elif len(astrosource_S) > 0:
        target = astrosource_S.pop()
    else:
        return np.nan, np.nan, np.nan, np.nan
  
    fwhm_L = list()

    for redf in redfL:
        try:
            xycen = centroid_quadratic(redf.mdata, *target.coord.to_pixel(redf.wcs), (15,15), search_boxsize=(5,5))
        except Exception as e:
            logger.warning(f"ReducedFit {redf.id} {target.name}: centroid_quadatric failed, using target.coord.to_pixel(redf.wcs), probably wrong")
            xycen = target.coord.to_pixel(redf.wcs)

        if not all(np.isfinite(xycen)):
            xycen = target.coord.to_pixel(redf.wcs)

        try:
            fwhm = fit_fwhm(xycen, redf=redf)[0]
            logger.debug(f"{target.name}: Gaussian FWHM: {fwhm:.1f} px")
            fwhm_L.append(fwhm)
        except Exception as e:
            logger.warning(f"ReducedFit {redf.id} {target.name}: error in gaussian fit, skipping this reduced fit")

    if len(fwhm_L) > 0:
        mean_fwhm = np.mean(fwhm_L)
    else:
        logger.error(f"Could not find an appropriate aperture for Reduced Fits {[redf.id for redf in redfL]}, using standard fwhm of 3.5px")
        mean_fwhm = 3.5

    sigma = mean_fwhm / (2*np.sqrt(2*math.log(2)))
    r = sigma
    
    return 5.0*r, 7.0*r, 15.0*r, {'mean_fwhm':mean_fwhm, 'sigma':sigma}


def fit_fwhm(pos_px: (float,float), data: NDArray = None, redf: 'ReducedFit' = None, px_max: int = None) -> float:
    r""" Fits a 1D gaussian + constant to the radial profile of the data around the given position, and returns the FWHM of the gaussian."""

    import numpy as np
    from photutils.profiles import RadialProfile
    from photutils.centroids import centroid_quadratic
    from astropy.modeling import models, fitting
    from iop4lib.instruments import Instrument
    from astropy.modeling.models import Moffat1D, Const1D, Gaussian1D

    if data is None:
        data = redf.mdata

    if px_max is None:
        # 0.4 arcsecs is excellent seeing, so let's consider 30 times that as the maximum radius
        px_max = int((30*0.4) / Instrument.by_name(redf.instrument).arcsec_per_pix)
        # this gives ~ 90 px for DIPOL, 30px for AndorT90

    pxs = np.arange(px_max)
    rp = RadialProfile(data, pos_px, pxs)

    fit = fitting.LevMarLSQFitter()

    # moffat = Moffat1D(x_0=0, amplitude=max(rp.profile)) + Const1D(min(rp.profile))
    # moffat[0].x_0.fixed = True
    # moffat_fit = fit(moffat, rp.radius, rp.profile)

    gaussian = Gaussian1D(amplitude=max(rp.profile), mean=0, stddev=1) + Const1D(min(rp.profile))
    gaussian[0].mean.fixed = True
    gaussian_fit = fit(gaussian, rp.radius, rp.profile)

    return gaussian_fit[0].fwhm

def fit_sigma(pos_px: (float, float), *args, **kwargs) -> float:
    r""" Fits a 1D gaussian + constant to the radial profile of the data around the given position, and returns the standard deviation of the gaussian."""
    fwhm = fit_fwhm(pos_px, *args, **kwargs)
    sigma = fwhm / (2*np.sqrt(2*math.log(2)))
    return sigma



def fit_gaussian(px_start, redf=None, data=None, sigma_start=7, r_max=None, r_search=None):
    r""" Fits a 2D gaussian + constant to the data around the given position, and returns the fitted model.
    
    Parameters
    -----------
        px_start: (float, float)
            Initial pixel position of the center of the gaussian.
        sigma_start: int (default 7)
            Initial guess for the sigma of the gaussian.
        r_max: int (default 90)
            Region radius around px_start in which to perform the fit.
        r_search: int or float (default None)
            If provided, it will search for the maximum in a circle of radius r_search around px_start, and use that as the starting point for the fit.
    """

    import numpy as np
    from photutils.profiles import RadialProfile
    from astropy.modeling import fitting
    from iop4lib.instruments import Instrument
    from astropy.modeling.models import Const2D, Gaussian2D
    from iop4lib.instruments import Instrument

    if redf is not None:
        if data is None:
            data = redf.mdata

        if r_max is None:
            # 0.4 arcsecs is excellent seeing
            r_max = int((30*0.4) / Instrument.by_name(redf.instrument).arcsec_per_pix) 
    else:
        if r_max is None:
            r_max = 90

    height, width = data.shape

    x_start, y_start = px_start

    X, Y = np.meshgrid(np.arange(width), np.arange(height))
    
    if r_search is not None:
        idx_region = np.sqrt((X-x_start)**2 + (Y-y_start)**2) < r_search
        idx_region_max = np.argmax(data[idx_region])
        x_start, y_start = X[idx_region][idx_region_max], Y[idx_region][idx_region_max]

    idx_fit_region = np.sqrt((X-x_start)**2 + (Y-y_start)**2) < r_max

    X = X[idx_fit_region].flatten()
    Y = Y[idx_fit_region].flatten()
    Z = np.ma.array(data[idx_fit_region]).compressed()

    fit = fitting.LevMarLSQFitter()
    gaussian = Gaussian2D(amplitude=data[int(y_start), int(x_start)], x_mean=x_start, y_mean=y_start, x_stddev=sigma_start, y_stddev=sigma_start) + Const2D(np.median(Z))
    gaussian[0].x_stddev.tied = lambda model: model[0].y_stddev
    gaussian_fit = fit(gaussian, X, Y, Z)

    return gaussian_fit


def get_angle_from_history(redf: 'ReducedFit' = None, 
                           target_src: 'AstroSource' = None, 
                           calibrated_fits: Sequence['ReducedFit'] = None, 
                           n_max_fits=20) -> Union[float, float]:
    """ Compute the average rotation angle from a list of already calibrated fits.

    To compute it, it checks the WCS in the header of each calibrated fit.

    If no list of calibrated fits is given, but a reduced fit and a target source 
    are given, it will try to get it existing and calibrated reduced fits in the DB 
    for the same instrument and target source, but considering only photometry images.
    It will use at most n_max_fits (default 20) reduced fits, but will try to get the 
    ones closer in time to the given reduced fit. If a list of reduced fits is given,
    it will use that list instead of querying the DB, and n_max_fits will be ignored.

    If no target source is given, it will use the header_hintobject of the reduced fit.
    """
    from iop4lib.enums import IMGTYPES, OBSMODES
    from iop4lib.db import ReducedFit
    from iop4lib.instruments import Instrument

    if calibrated_fits is None:
        if target_src is None:
            target_src = redf.header_hintobject

        qs = ReducedFit.objects.filter(instrument=redf.instrument, 
                                                imgtype=IMGTYPES.LIGHT, 
                                                obsmode=OBSMODES.PHOTOMETRY, 
                                                flags__has=ReducedFit.FLAGS.BUILT_REDUCED,
                                                sources_in_field__in=[target_src])
        
        if len(qs) == 0:
            logger.warning(f"No calibrated fits for {redf.instrument} {target_src.name}, using other sources too")
            qs = ReducedFit.objects.filter(instrument=redf.instrument, 
                                                imgtype=IMGTYPES.LIGHT, 
                                                obsmode=OBSMODES.PHOTOMETRY, 
                                                flags__has=ReducedFit.FLAGS.BUILT_REDUCED)
        if len(qs) < 5:
            logger.warning(f"Less than 5 calibrated fits for {redf.instrument} {target_src.name}, using all of them")

        jds = np.array(qs.values_list('juliandate', flat=True))
        idx = np.argsort(np.abs(jds - redf.juliandate))[0:n_max_fits]
        calibrated_fits = [qs[int(i)] for i in idx]

    angle_L = list()
    for calibrated_fit in calibrated_fits:
        try:
            w = WCS(calibrated_fit.header, key="A")
        except Exception as e: # usually KeyError
            logger.warning(f"Could not get WCS 'A' from ReducedFit {calibrated_fit.id}: {e}")
            continue
            
        
        # Extract the PC matrix elements
        pc_11, pc_12 = w.wcs.pc[0]
        pc_21, pc_22 = w.wcs.pc[1]
        
        # Calculate the rotation angle in degrees
        angle = np.degrees(np.arctan2(pc_21, pc_11)) % 360 # wrap angles at 0,360 (-180 = 180)

        if 'FLIPSTAT' in redf.rawfit.header and redf.rawfit.header['FLIPSTAT'] == 'Flip':
                angle = - angle

        angle_L.append(angle)

    if len(angle_L) == 0:
        logger.warning(f"No angle found in history for {redf.instrument} {target_src.name}")

        if hasattr(Instrument.by_name(redf.instrument), "default_sky_angle"):
            logger.warning(f"Using default sky angle for {redf.instrument}")
            return Instrument.by_name(redf.instrument).default_sky_angle, Instrument.by_name(redf.instrument).default_sky_angle_std
        else:
            logger.error(f"No default sky angle for {redf.instrument}, returning NaN")
            return np.nan, np.nan
    
    angle_mean = np.mean(angle_L)
    angle_std = np.std(angle_L)

    return angle_mean, angle_std

def build_wcs_centered_on(target_px: (float, float),
                          target_coord: (float, float) = None,
                          target_src : 'AstroSource' = None, 
                          redf: 'ReducedFit' = None, 
                          angle: float = None, 
                          pixel_scale: float = None) -> WCS:
    r""" Build a WCS object with the target source at the target pixel position.

    Builds a WCS object that has the target astrosource at the given pixel position in the image,
    with the given angle (in degrees) and pixel scale (in degrees/pixel).

    If either angle, pixel_scale or target source is not given, it will try to get them
    from the given reduced fit.

    If no pixel scale is given, it will use the pixel scale of the instrument of the 
    reduced fit.

    If no angle is given, it will try to get it from the history of the reduced fits for the same 
    instrument and target source, but considering only photometry images.

    If no target source is given, it will use the header_hintobject of the reduced fit.
    """

    from iop4lib.instruments import Instrument

    if angle is None:
        angle, angle_std = get_angle_from_history(redf, target_src) # in degrees
        if angle_std > 0.5:
            logger.warning(f"Large angle std: {angle=}, {angle_std=}")
    
    if pixel_scale is None:
        pixel_scale = Instrument.by_name(redf.instrument).arcsec_per_pix / 3600 # in degrees

    if target_src is None and redf is not None:
        target_src = redf.header_hintobject

    if target_coord is None and target_src is not None:
        known_ra, known_dec = target_src.coord.ra.deg, target_src.coord.dec.deg, 
    else:
        known_ra, known_dec = target_coord.ra.deg, target_coord.dec.deg

    known_x_pixel, known_y_pixel = target_px[0], target_px[1]

    w = WCS(naxis=2)

    # Set RA, DEC for reference pixel (in degrees)
    w.wcs.crval = [known_ra, known_dec]

    # Set reference pixel
    w.wcs.crpix = [known_x_pixel, known_y_pixel] 

    # Other settings remain the same
    w.wcs.cdelt = np.array([pixel_scale, pixel_scale])
    w.wcs.pc = [[+np.cos(np.radians(angle)), -np.sin(np.radians(angle))],
                [+np.sin(np.radians(angle)), +np.cos(np.radians(angle))]]

    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]

    return w



@dataclasses.dataclass
class SimbadSource():
    name: str
    ra_hms: str
    dec_dms: str
    otype: str
    other_names: str = None


    def __str__(self):
        return f"<Simbad source {self.name}>"

    @property
    def coord(self):
        return SkyCoord(self.ra_hms, self.dec_dms, unit=(u.hourangle, u.deg))
    
    def is_in_field(self, wcs, height, width):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                x, y = self.coord.to_pixel(wcs)
        except:
            return False
        else:
            if (0 <= x < height) and (0 <= y < width):
                return True
            else:
                return False
            

def get_simbad_table(center_coord, radius, Nmax=6, all_types=False):
    from astroquery.simbad import Simbad
    Simbad.ROW_LIMIT = 100
    Simbad.add_votable_fields('otype')
    Simbad.add_votable_fields('flux(R)', 'flux(G)')

    tb_simbad = Simbad.query_region(center_coord, radius=radius, epoch="J2000")

    if not all_types:
        otypes = ["Star", "QSO", "BLLac", "BLLac"]

        idx = np.full(len(tb_simbad), fill_value=False)
        for otype in otypes:
            idx = idx | (tb_simbad["OTYPE"] == otype)

        tb_simbad = tb_simbad[idx]

    if len(tb_simbad) > Nmax:
        tb_simbad = tb_simbad[0:Nmax]

    return tb_simbad

def get_simbad_sources(center_coord, radius, Nmax=6, all_types=False, exclude_self=True):

    tb_simbad = get_simbad_table(center_coord, radius, Nmax=6, all_types=all_types)

    simbad_sources = list()

    for row in tb_simbad:
        name = row['MAIN_ID']
        coord = SkyCoord(Angle(row['RA'], unit=u.hourangle), Angle(row['DEC'], unit=u.degree), frame='icrs')
        ra_hms = coord.ra.to_string(unit=u.hourangle, sep=':', precision=4, pad=True)
        dec_dms = coord.dec.to_string(unit=u.degree, sep=':', precision=4, pad=True)
        otype = row['OTYPE']

        simbad_sources.append(SimbadSource(name=name, ra_hms=ra_hms, dec_dms=dec_dms, otype=otype))

    if exclude_self:
        simbad_sources = [simbad_source for simbad_source in simbad_sources if simbad_source.coord.separation(center_coord).arcsec > 1]

    return simbad_sources




def get_host_correction(astrosource, aperas, fwhm=None) -> (float, float):
    r""" Returns the contaminating flux and its uncertainty for a given astrosource and aperture radius. 
    
    If no correction is available for the given astrosource, it returns None, None.

    The value is interpolated from the tables in `host_correction_data.csv`. You can find
    more details in this file.

    Parameters
    ----------
    astrosource: AstroSource
        The astrosource for which to get the correction.
    aperas: float
        The aperture radius in arcsecs.
    fwhm: float, optional
        The fwhm in arcsec. Default is 7 arcsec (approximately this 
        fwhm for our 0.9m to 2.2m telescopes, doesnt affect much)

    Returns
    -------
    flux, flux_err: tuple[float, float]
        The contaminating flux and its uncertainty in mJy.

    or
    
    None, None

    """

    import numpy as np
    import pandas as pd
    from io import StringIO
    from importlib import resources
    import re

    with open(resources.files("iop4lib.utils") / "host_correction_data.csv", 'r') as f:
        text = f.read()

    df = None

    def get_invariable_str(s):
        return s.replace(' ', '').replace('-','').replace('+','').replace('_','').upper()
    
    for table in text.split("#"*80):
        objname = re.findall(r"OBJECT: (.*)", table)[0]
        if get_invariable_str(astrosource.name) == get_invariable_str(objname) or \
            (astrosource.other_names is not None and any([get_invariable_str(other_name) == get_invariable_str(objname) for other_name in astrosource.other_names_list])):
            df = pd.read_csv(StringIO(table), comment="#")
            break

    if df is None:
        return None, None
    
    if fwhm is None:
        fwhm = 7 

    # Get the column whose header is closest to the indicated fwhm
    
    fwhm_L = np.array([float(fwhm) for fwhm in df.columns[1:]])
    fwhm_idx = np.argmin(np.abs(fwhm_L - fwhm))

    # Get the values and uncertainties of the contaminating flux for the indicated fwhm

    flux_L, flux_err_L = zip(*[v.split(" ") for v in df.iloc[:,fwhm_idx+1]])

    flux_L = np.array([float(flux) for flux in flux_L])
    flux_err_L = np.array([float(flux_err) for flux_err in flux_err_L])

    aperas_L = np.array([float(apera) for apera in df.iloc[:,0]])

    # Interpolate the contaminating flux and its uncertainty for the indicated aperture radius
    # Use simple linear interpolation from np.interp:

    flux = np.interp(aperas, aperas_L, flux_L)
    flux_err = np.interp(aperas, aperas_L, flux_err_L)

    return flux, flux_err



# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# other imports
import os
import glob
from pathlib import Path
import astropy.io.fits as fits
import psutil
import multiprocessing
import time
import time
import signal
from logging.handlers import QueueHandler
import traceback

# iop4lib imports
from iop4lib.utils import  get_mem_parent_from_child, get_total_mem_from_child, get_mem_current, get_mem_children

# logging
import logging
logger = logging.getLogger(__name__)


# BUILD REDUCED FILES IN MULTIPROCESSING POOL

def epoch_bulkreduce_multiprocesing(reduced_L, epoch=None):
    """ Reduces a list of ReducedFit instances in a multiprocessing pool.

    Invokes the reduction of a list of ReducedFit instances in a multiprocessing pool, with a maximum number of
    concurrent processes defined by iop4conf.nthreads. It can be invoked with a list of ReducedFit 
    from different epochs.

    Parameters
    ----------
    reduced_L : list of ReducedFit
        List of ReducedFit instances to be reduced.

    Other Parameters
    ----------------
    epoch : Epoch, optional
        If provided, it is used only to print the epoch in the log messages of the main thread.
    """

    logger.info(f"{epoch}: starting {iop4conf.nthreads} threads to build {len(reduced_L)} reduced files. Current memory usage: {get_mem_current()/1024**3:.2f} GB.")

    mp_ctx = multiprocessing.get_context('spawn')

    queue = mp_ctx.Manager().Queue(-1)  # create queue listener for multiprocessing logging (-1 so no limit on size)
    listener = logging.handlers.QueueListener(queue, *logging.getLogger().handlers)
    listener.start()
    
    counter = mp_ctx.Value('i', 0)

    with mp_ctx.Pool(processes=iop4conf.nthreads, 
                              initializer=_epoch_bulkreduce_multiprocessing_init,
                              initargs=(counter, queue, len(reduced_L), iop4conf), 
                              maxtasksperchild=20) as pool:

        tasks = [pool.apply_async(_epoch_bulkreduce_multiprocessing_worker, args=(redf,)) for redf in reduced_L]    
    
        pool.close() # no more tasks to be submitted

        _epoch_bulkreduce_multiprocesing_mainloop(tasks, counter, reduced_L, epoch=epoch)
        
        pool.join() # wait for all tasks to finish (should already be finished anyway)
        pool.terminate()
    
    listener.stop()

    logger.info(f"{epoch}: all threads finished. Current memory usage: {get_mem_current()/1024**3:.2f} GB.")



def _epoch_bulkreduce_multiprocessing_init(counter, queue, Nredf, iop4conf):
    """ helper func of bulkreduce_multiprocessing to configure the child processes.
    
    Configures the root logger to send messages to a common queue and to allow only messages 
    from iop4lib (and not from the astrometry solver). iopconf has to be passed as an argument, 
    as if method is spawn, the debug level indicated in the cli is not kept.
    
    Makes _counter (shared) and _Nredf global variables so they can be used by the child processes to keep track of the progress.

    Sets up django ORM so it is usable from the child. 
    """

    # counter and total to show progress

    global _counter
    _counter = counter

    global _Nredf
    _Nredf = Nredf

    # set up logging to send messages to a common queue

    h = QueueHandler(queue)    

    root = logging.getLogger()
    root.handlers.clear()
    iop4conf.configure_logging() # force configure logging 
    root.addHandler(h)
    root.addFilter(logging.Filter("iop4lib"))

    root.setLevel(iop4conf.log_level)

    logger = logging.getLogger(__name__)

    # set up django ORM to be available

    import iop4lib
    iop4conf = iop4lib.Config(**iop4conf)

    # fix for linux where the settings keep being configured even with spawn method
    from django.conf import settings
    if not settings.configured:
        iop4conf.configure_db()

    # close existing connections to the DB
    from django import db
    db.connections.close_all()


class TimeoutException(Exception):
    pass



def _epoch_bulkreduce_multiprocessing_worker_timeout_handler(signum, frame):
    raise TimeoutException



def _epoch_bulkreduce_multiprocessing_worker(reduced_fit: 'ReducedFit'):
    """ helper func to invoke .build() method of ReducedFit on instances from a multiprocessing pool. 

    It needs to be a top-level function, not a method of the class, so it can be pickled and sent to 
    the pool.
    """
    from iop4lib.db import ReducedFit

    logger = logging.getLogger(__name__)
    
    logger.debug(f"Epoch {reduced_fit.epoch.id}, child pid {os.getpid()}, ReducedFit {reduced_fit.id} (parent: {get_mem_parent_from_child()/1024**2:.0f} MB, this child: {get_mem_current()/1024**2:.0f} MB, total: {get_total_mem_from_child()/1024**3:.2f} GB total): started building file {_counter.value} / {_Nredf} ({100*_counter.value/_Nredf:.0f}%).")


    try:
        # Start a timer that will send SIGALRM in iop4conf.astrometry_timeout minutes
        signal.signal(signal.SIGALRM, _epoch_bulkreduce_multiprocessing_worker_timeout_handler)
        signal.alarm(iop4conf.astrometry_timeout*60) 
        reduced_fit.build_file()
        signal.alarm(0) # cancel the alarm
    except Exception as e:
        logger.error(f"Epoch {reduced_fit.epoch.id}, child pid {os.getpid()}, ReducedFit {reduced_fit.id}: ERROR. Exception was: {e}. Traceback: {traceback.format_exc()}")
    else:
        with _counter.get_lock():
            _counter.value += 1
        logger.debug(f"Epoch {reduced_fit.epoch.id}, child pid {os.getpid()}, ReducedFit {reduced_fit.id}: finished building file {_counter.value} / {_Nredf} ({100*_counter.value/_Nredf:.0f}%).")





def _epoch_bulkreduce_multiprocesing_mainloop(tasks, counter, reduced_L, epoch=None):

    time_start = time.time()

    while not all([task.ready() for task in tasks]):

        time_elapsed = time.time() - time_start
        avg_pace = time_elapsed / counter.value if counter.value > 0 else float('inf')

        try:
            childs = [{'pid':child.pid, 'mem':child.memory_info().rss} for child in psutil.Process().children(recursive=True)]
        except psutil.NoSuchProcess:
            # one of the child processes might have dissapeared or restarted in the meantime (because of maxtasksperchild or 
            # because the tasks are close to being finished), so we just ignore it.
            pass
        else:
            logger.info(
                                "\n" + "*"*40 + "\n" +
                                "\n" +
                                (f"[ {epoch} - Main Thread (pid {os.getpid()}) ]\n" if epoch is not None else f"[ Main Thread (pid {os.getpid()}) ]\n") +
                                "\n" +
                                f"Progress: {counter.value} of {len(reduced_L)} files processed in {time_elapsed:.0f} seconds ({avg_pace:.1f} s / file).\n" 
                                "\n" +
                                f"Parent memory: {get_mem_current()/1024**2:5.0f} MB\n" +
                                f"Children memory: {get_mem_children()/1024**3:.2f} GB\n" +
                                "\n".join([f"  - pid {child['pid']:5} : {child['mem']/1024**2:5.0f} MB" for child in childs]) + "\n" +
                                "\n" + 
                                "*"*40
                            )

        time.sleep(5)

    # When task have finished, return
    return


# COMPUTE RELATIVE POLARIMETRY IN MULTIPROCESSING POOL

def _parallel_relative_polarimetry_helper(keys, group):
    from iop4lib.instruments import Instrument
    try:
        Instrument.by_name(keys['instrument']).compute_relative_polarimetry(group)
    except Exception as e:
        logger.error(f"Error for {group=}. Exception {e}: {traceback.format_exc()}")
    finally:
        logger.info(f"Finished computing relative polarimetry for {group=}.")

def parallel_relative_polarimetry(keys, groups):
    mp_ctx = multiprocessing.get_context('fork') # fork is faster and does not need configuring again
    with mp_ctx.Pool(iop4conf.nthreads) as pool:
        pool.starmap(_parallel_relative_polarimetry_helper, zip(keys, groups))


"""
Some plotting utilities for the IOP4 project.
"""

import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)  

# other imports

import numpy as np
import scipy as sp
import matplotlib as mplt
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import itertools
from astropy.coordinates import Angle, SkyCoord
from astropy.wcs import WCS
import astropy.units as u
from astropy.visualization import SqrtStretch, LogStretch, AsymmetricPercentileInterval
from astropy.visualization.mpl_normalize import ImageNormalize
from photutils.aperture import CircularAperture
from pathlib import Path

from iop4lib.utils.sourcepairing import (get_pairs_d, get_pairs_dxy, get_best_pairs)

from .sourcedetection import select_points

# logging

import logging
logger = logging.getLogger(__name__)

def hist_data(data, log=True, ax=None):
                
    if ax is None:
        ax = plt.gca()

    if log:
        ax.set_xscale('log')
        ax.set_yscale('log')

    ax.hist(data, bins="sqrt", log=log, histtype='step')
    
    ax.axvline(x=np.quantile(data, 0.3), color="r")
    ax.axvline(x=np.quantile(data, 0.5), color="g")
    ax.axvline(x=np.quantile(data, 0.99), color="b") 

    ax.xaxis.set_major_locator(mplt.ticker.MaxNLocator(4))

def imshow_w_sources(imgdata, pos1=None, pos2=None, normtype="log", vmin=None, vmax=None, a=10, cmap=None, ax=None, r_aper=20):
    if ax is None:
        ax = plt.gca()

    if isinstance(imgdata, np.ma.masked_array):
        data = imgdata.compressed()
    else:
        data = np.array(imgdata[np.isfinite(imgdata)]).flatten()

    if cmap is None:
        cmap = plt.cm.gray
        cmap.set_bad(color='red')
        cmap.set_under(color='black')
        cmap.set_over(color='white')
            
    if vmin is None:
        vmin = np.quantile(data, 0.3)

    if vmax is None:
        vmax = np.quantile(data, 0.99)

    if normtype == "log":
        #norm = ImageNormalize(imgdata, interval=AsymmetricPercentileInterval(30, 99), stretch=LogStretch(a=10))
        norm = ImageNormalize(imgdata, vmin=vmin, vmax=vmax, stretch=LogStretch(a=a))
        #norm = LogNorm(vmin=vmin, vmax=vmax)
    elif normtype == "logstretch":
        norm = ImageNormalize(stretch=LogStretch(a=a))
    elif normtype == "sqrtstretch":
        norm = ImageNormalize(stretch=SqrtStretch())
    else:
        pass

    ax.imshow(imgdata, cmap=cmap, origin='lower', norm=norm)    

    pos1_present = pos1 is not None and len(pos1) > 0
    pos2_present = pos2 is not None and len(pos2) > 0

    if pos1_present and not pos2_present:
        apertures1 = CircularAperture(pos1, r=r_aper)
        apertures1.plot(color="r", lw=1, alpha=0.9, linestyle='--', ax=ax)
            
    if pos1_present and pos2_present:

        if len(pos1) < 300:
            apertures1 = CircularAperture(pos1, r=r_aper)
            apertures2 = CircularAperture(pos2, r=r_aper)
            
            color_cycle = itertools.cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])
            colors = [next(color_cycle) for _ in range(len(apertures1))]

            for i, (ap1, ap2) in enumerate(zip(apertures1, apertures2)):
                ap1.plot(color=colors[i], lw=1, alpha=0.9, linestyle='--', ax=ax)
                ap2.plot(color=colors[i], lw=1, alpha=0.9, linestyle='-', ax=ax)
        else:
            apertures1 = CircularAperture(pos1, r=r_aper)
            apertures2 = CircularAperture(pos2, r=r_aper)
            
            apertures1.plot(color="m", lw=1, alpha=0.9, linestyle='--', ax=ax)
            apertures2.plot(color="y", lw=1, alpha=0.9, linestyle='-', ax=ax)



def plot_preview_background_substraction_1row(redf, bkg, axs=None, fig=None):

    imgdata_bkg_substracted = redf.mdata - bkg.background

    if axs is None:
        if fig is None:
            fig = plt.gcf()
        axs = fig.subplots(nrows=1, ncols=4)

    axs[0].set_title('Original')
    imshow_w_sources(redf.mdata, ax=axs[0])
    axs[1].set_title('Background')
    imshow_w_sources(bkg.background, ax=axs[1])
    axs[2].set_title('Background-subtracted Data')
    imshow_w_sources(imgdata_bkg_substracted, ax=axs[2])
    axs[3].set_title('Background RMS')
    imshow_w_sources(bkg.background_rms, ax=axs[3])

    
def plot_preview_background_substraction(redf, bkg, axs=None, fig=None):

    if axs is None:
        if fig is None:
            fig = plt.gcf()
        axs = fig.subplots(nrows=2, ncols=4)

    imgdata_bkg_substracted = redf.mdata - bkg.background

    axs[0,0].set_title('Original')
    imshow_w_sources(redf.mdata, ax=axs[0,0])
    hist_data(redf.mdata.compressed(), ax=axs[1,0])
    axs[0,1].set_title('Background')
    imshow_w_sources(bkg.background, ax=axs[0,1])
    hist_data(bkg.background.flatten(), ax=axs[1,1])
    axs[0,2].set_title('Background-subtracted Data')
    imshow_w_sources(imgdata_bkg_substracted, ax=axs[0,2])
    hist_data(imgdata_bkg_substracted.compressed(), ax=axs[1,2])
    axs[0,3].set_title('Background RMS')
    imshow_w_sources(bkg.background_rms, ax=axs[0,3])
    hist_data(bkg.background_rms.flatten(), ax=axs[1,3])


def plot_preview_astrometry(redf, with_simbad=False, legend=True, names_over=False, ax=None, fig=None, **astrocalib_proc_vars):

    from iop4lib.db import AstroSource
    from iop4lib.utils import get_simbad_sources

    def get_matched_sources(match, pos, d_eps=1.412):
        """ Finds sources matching field stars within sqrt(2) pixels. """
        
        wcs = WCS(match.wcs_fields)
        raL = [star.ra_deg*u.deg for star in match.stars]
        decL = [star.dec_deg*u.deg for star in match.stars]
        
        match_coords = SkyCoord(ra=raL, dec=decL)
        matchs_pix = wcs.world_to_pixel(match_coords)

        matched_stars = list()
        
        for i, pix in enumerate(zip(*matchs_pix)):
            d = np.sqrt((pix[0]-pos[:,0])**2 + (pix[1]-pos[:,1])**2)
            if np.min(d) < d_eps:
                matched_stars.append(match_coords[i])
                
        return matched_stars

    ## if wcs1, wcs2 in astrocalib_proc_vars, use them, otherwise use the ones in redf

    if 'wcs1' in astrocalib_proc_vars:
        wcs1 = astrocalib_proc_vars['wcs1']
    else:
        wcs1 = redf.wcs1

    has_pairs = astrocalib_proc_vars.pop('has_pairs', redf.has_pairs)
    
    if has_pairs:
        if 'wcs2' in astrocalib_proc_vars:
            wcs2 = astrocalib_proc_vars['wcs2']
        else:
            wcs2 = redf.wcs2

    ## get detected stars sources in the image
    if 'stars' in astrocalib_proc_vars and astrocalib_proc_vars['stars'] is not None:
        detected_stars = astrocalib_proc_vars['stars']
    else:
        detected_stars = None

    ## get the index stars from the astrometry match 
    if 'solution' in astrocalib_proc_vars:
        indexstars_ra, indexstars_dec = list(zip(*[[star.ra_deg, star.dec_deg] for star in astrocalib_proc_vars['solution'].best_match().stars]))
        indexstars_coords = SkyCoord(ra=indexstars_ra*u.deg,  dec=indexstars_dec*u.deg)
    else:
        indexstars_coords = None

    ## get index stars matching within sqrt(2) pixels of the used centroids.
    if 'bm' in astrocalib_proc_vars and 'stars' in astrocalib_proc_vars:
        matched_stars = get_matched_sources(astrocalib_proc_vars['bm'], np.array(astrocalib_proc_vars['stars']))
    else: 
        matched_stars = None

    ## get sources in field
    sources_in_field = AstroSource.get_sources_in_field(wcs1, width=redf.width, height=redf.height)
    logger.debug(f"{redf}: found {len(sources_in_field)} catalog sources in field: {sources_in_field}")


    if fig is None:
        fig = plt.gcf()

    if ax is None:
        if len(fig.axes) > 0:
            ax = plt.gca()
        else:
            ax = fig.add_subplot(projection=redf.wcs)

    legend_handles_L = list()
    legend_labels_L = list()

    # image

    imshow_w_sources(redf.mdata, ax=ax)

    # # detected stars

    # if detected_stars is not None:
    #     ap = CircularAperture(detected_stars, r=20.0)
    #     h = ap.plot(color="g", lw=1, alpha=1.0, linestyle='--', ax=ax)[-1]
    #     legend_handles_L.append(h)
    #     legend_labels_L.append("detected sources")

    # # index stars

    # if indexstars_coords is not None:
    #     aps = CircularAperture(list(zip(*SkyCoord(indexstars_coords).to_pixel(wcs1))), r=20)
    #     h = aps.plot(color="b", lw=1, alpha=0.8, linestyle='-', ax=ax)[-1]
    #     legend_handles_L.append(h)
    #     legend_labels_L.append("index stars")

    # # matched stars

    # if matched_stars is not None and len(matched_stars) > 0:
    #     aps = CircularAperture(list(zip(*SkyCoord(matched_stars).to_pixel(wcs1))), r=20)
    #     h = aps.plot(color="y", lw=1, alpha=0.8, linestyle='--', ax=ax, label="matched stars")[-1]
    #     legend_handles_L.append(h)
    #     legend_labels_L.append("matched stars (d<sqrt(2))")

    # catalogue sources

    if sources_in_field is not None and len(sources_in_field) > 0:
        for i, source in enumerate(sources_in_field):
            ap = CircularAperture([*source.coord.to_pixel(wcs1)], r=20)
            h = ap.plot(color="r", lw=1, alpha=1, linestyle='-', ax=ax, label=f"{source.name}")
            if has_pairs:
                ax.plot(*source.coord.to_pixel(wcs2), 'rx', alpha=1)
            x, y = source.coord.to_pixel(wcs1)
            ax.annotate(text=source.name if names_over else f"{i}", 
                        xy=(x+20, y), 
                        xytext=(40,0) if names_over else (10,0), 
                        textcoords="offset pixels", 
                        color="red", fontsize=10, weight="bold",
                        verticalalignment="center",
                        #path_effects=[mplt.patheffects.withStroke(linewidth=0.5, foreground="black")],
                        arrowprops=dict(color="red", width=0.5, headwidth=1, headlength=3))
            
            if not names_over:
                legend_handles_L.append(h)
                legend_labels_L.append(f"{i}: {source.name}")

        if names_over:
            legend_handles_L.append(h)
            legend_labels_L.append(f"Catalog sources")

    # overlay sources in Simbad

    if with_simbad:
        try:
            # known sources in this area of sky (Simbad)
            pixscale = np.mean(np.sqrt(np.sum(wcs1.pixel_scale_matrix**2, axis=0)))*u.Unit("deg/pix")
            radius = redf.width * u.Unit('pixel') * pixscale
            centercoord = wcs1.pixel_to_world(redf.mdata.shape[0]//2, redf.mdata.shape[1]//2)

            simbad_sources = get_simbad_sources(centercoord, radius, Nmax=6, exclude_self=False)

            for i, src in enumerate(simbad_sources):
                x, y = src.coord.to_pixel(wcs1)
                ap = CircularAperture([x,y], r=20)
                h = ap.plot(color="yellow", lw=1, alpha=0.8, linestyle='--', ax=ax, label=src.name)
                ax.annotate(text=src.name if names_over else f"{i}", 
                            xy=(x+20, y) if names_over else (x-20, y), 
                            xytext=(40,-20) if names_over else (-15,0), 
                            verticalalignment="center",
                            textcoords="offset pixels", color="yellow", fontsize=10, weight="bold", arrowprops=dict(color='yellow', width=1.0, headwidth=1, headlength=3) if names_over else None)
                if not names_over:
                    legend_handles_L.append(h)
                    legend_labels_L.append(f"{i}: {src.name}")
       
        except Exception as e:
            logger.debug(f"Simbad query failed, ignoring: {e}")

    # Plot the displacement betwen pairs as scale, if specified

    if (disp_sign := astrocalib_proc_vars.get('disp_sign', None)) is not None:
        ax.arrow(x=redf.mdata.shape[0]-20, y=20, dx=disp_sign[0], dy=disp_sign[1], color='red', lw=2, head_width=8, length_includes_head=True)

    # set axes and legends

    if legend:
        if names_over:
            ax.legend(legend_handles_L, legend_labels_L, facecolor=(1,1,1,0.5))
        else:
            ax.legend(legend_handles_L, legend_labels_L, title="Sources", facecolor=(1,1,1,0.5), loc="lower left", bbox_to_anchor=(1,0))

    ax.set_xlim([0, redf.mdata.shape[1]])
    ax.set_ylim([0, redf.mdata.shape[0]])
    ax.coords.grid(True, color='white', ls='solid')
    ax.coords[0].set_axislabel_visibility_rule("always")
    ax.coords[1].set_axislabel_visibility_rule("always")






    



def photopolresult_mplt_viewer(src, band="R", qs0 = None):
    from iop4lib.db import PhotoPolResult, AstroSource
    from astropy.time import Time
    from django.db.models import Q, F, Func, Avg, StdDev, Min, Max

    if isinstance(src, str):
        astrosource = AstroSource.objects.get(name=src)
        srcname = src
    elif isinstance(src, AstroSource):
        astrosource = src
        srcname = astrosource.name
    else:
        raise ValueError("src must be either a string or an AstroSource object")
    
    if qs0 is None:
        qs0 = PhotoPolResult.objects.order_by('-juliandate')

    telescopes = ["OSN-T090", "OSN-T150", "CAHA-T220"]
    colors = ['b', 'g', 'r']

    fig, axs = plt.subplots(nrows=3, figsize=(10,10), sharex=True)
        
    lines_L = list()

    for telescope, color in zip(telescopes, colors):
        qs = qs0.filter(astrosource__name=srcname, epoch__telescope=telescope, band=band)
        
        if qs.count() == 0:
            continue
            
        values_lists = zip(*qs.values_list('juliandate', 'mag', 'mag_err', 'p', 'p_err', 'chi', 'chi_err', 'pk'))
        
        values_lists = [[x if x is not None else np.nan for x in values] for values in values_lists]
        jd, mag, mag_err, p, p_err, chi, chi_err, pk = map(np.array, values_lists)
        datetime = Time(jd, format="jd").datetime

        lines_L.append((0, pk, axs[0].errorbar(x=datetime, y=mag, yerr=mag_err, marker=".", color=color, linestyle="none").lines[0]))
        lines_L.append((1, pk, axs[1].errorbar(x=datetime, y=p, yerr=p_err, marker=".", color=color, linestyle="none").lines[0]))
        lines_L.append((2, pk, axs[2].errorbar(x=datetime, y=chi, yerr=chi_err, marker=".", color=color, linestyle="none").lines[0]))

    # invert magnitude axis
    axs[0].invert_yaxis()

    # secondary axes and its x limits
    datetime_min, datetime_max = mplt.dates.num2date(axs[0].get_xlim())
    mjd_min, mjd_max = Time([datetime_min, datetime_max]).mjd
    ax0_2 = axs[0].twiny()
    ax0_2.set_xlim([mjd_min, mjd_max])

    # x and secondary x labels

    axs[-1].set_xlabel('date')
    ax0_2.set_xlabel('MJD')

    # y labels

    axs[0].set_ylabel(f"{band} mag")
    axs[1].set_ylabel("p")
    axs[1].yaxis.set_major_formatter(mplt.ticker.PercentFormatter(1.0, decimals=1))
    axs[2].set_ylabel("chi [º]")

    # title 
    fig.suptitle(f"{srcname} ({astrosource.other_names})")

    # legend
    legend_handles = [axs[0].plot([],[],color=color, marker=".", linestyle="none", label=telescope)[0] for color, telescope in zip(colors, telescopes)]
    legend_labels = telescopes
    fig.legend(legend_handles, legend_labels, loc="lower left", ncols=3)

    # hover
    qs = PhotoPolResult.objects.order_by('-juliandate').filter(astrosource__name=srcname)

    if qs.count() == 0:
        raise Exception
        
    annots = [ax.annotate("", xy=(0,0), xytext=(-20,20), textcoords="offset points", fontsize=8, bbox=dict(boxstyle="round", fc="gray", pad=0.5), arrowprops=dict(arrowstyle="->")) for ax in axs]

    for annot in annots:
        annot.set_visible(True)

    def update_annots(ax_i, event, line, ind, pks):

        idx = ind["ind"][0]
        pk = pks[idx] 
        obj = PhotoPolResult.objects.get(pk=pk)

        x = Time(obj.juliandate, format="jd").datetime
        y = getattr(obj, ['mag', 'p', 'chi'][ax_i])

        annot = annots[ax_i]
        annot.xy = [x, y]
        print([x,y])

        text = ((f"{obj.epoch.epochname}\n") + 
                (f"id {pk} / {obj.obsmode}\n") + 
                (f"{obj.band}  {obj.mag:.2f} $\\pm$ {obj.mag_err:.2f}\n" if obj.mag is not None else '') + 
                (f"p [%]  {100*obj.p:.1f} $\\pm$ {100*obj.p_err:.1f}\n" if obj.p is not None else '') + 
                (f"chi [º]  {obj.chi:.1f} $\\pm$ {obj.chi_err:.1f}" if obj.chi is not None else ''))
        
        import sys
        print(text, file=sys.__stdout__)
        
        annot.set_text(text)
        annot.get_bbox_patch().set_alpha(0.2)

    def on_plot_hover(event):
        try:
            annots_visible = [False for annot in annots]

            for i, pks, line in lines_L:
                cont, ind = line.contains(event)
                if cont:
                    update_annots(i, event, line, ind, pks)
                    annots_visible[i] = True

            for annot, visible in zip(annots,annots_visible):
                annot.set_visible(visible)

            fig.canvas.draw_idle()
        except Exception as e:
            pass
            #import sys, traceback
            #print(f"{e} {traceback.format_exc()}", file=sys.__stdout__)


    fig.canvas.mpl_connect('motion_notify_event', on_plot_hover)  

    return fig, axs






def build_astrometry_summary_images(redf, astrocalib_proc_vars, summary_kwargs):

    # Summary image of background substraction results

    logger.debug(f"{redf}: plotting astrometry summary image of background substraction results")

    fig = mplt.figure.Figure(figsize=(14,7), dpi=iop4conf.mplt_default_dpi)
    axs = fig.subplots(nrows=2, ncols=4)
    plot_preview_background_substraction(redf, astrocalib_proc_vars['bkg'], axs=axs)
    fig.savefig(Path(redf.filedpropdir) / "astrometry_1_bkgsubstraction.png", bbox_inches="tight")
    fig.clf()

    # Summary images of segmentation and pair finding results

    if 'pos_seg' in astrocalib_proc_vars.keys():

        logger.debug(f"{redf}: plotting astrometry summary image of segmentation results")

        if astrocalib_proc_vars['has_pairs']:
            fig = mplt.figure.Figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi)
            axs = fig.subplots(nrows=2, ncols=4)

            axs[0,0].set_title('Segmented Image')
            axs[0,0].imshow(astrocalib_proc_vars['segment_map'], origin="lower", cmap=astrocalib_proc_vars['segment_map'].cmap, interpolation="nearest")

            axs[1,0].set_title(f"Kron apertures (N={len(astrocalib_proc_vars['pos_seg'])})")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], ax=axs[1,0])
            astrocalib_proc_vars['seg_cat'].plot_kron_apertures(ax=axs[1,0], color='red', lw=1.5)

            axs[0,1].set_title(f"Pairs (n={len(astrocalib_proc_vars['seg1'])}, {len(astrocalib_proc_vars['seg1'])/len(astrocalib_proc_vars['pos_seg'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['seg1'], pos2=astrocalib_proc_vars['seg2'], ax=axs[0,1])

            axs[1,1].set_title(f"PairsXY (n={len(astrocalib_proc_vars['seg1xy'])}, {len(astrocalib_proc_vars['seg1xy'])/len(astrocalib_proc_vars['pos_seg'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['seg1xy'], pos2=astrocalib_proc_vars['seg2xy'], ax=axs[1,1])

            axs[0,2].set_title(f"BestPairs (n={len(astrocalib_proc_vars['seg1_best'])}, {len(astrocalib_proc_vars['seg1_best'])/len(astrocalib_proc_vars['pos_seg'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['seg1_best'], pos2=astrocalib_proc_vars['seg2_best'], ax=axs[0,2])

            axs[1,2].set_title(f"BestPairsXY (n={len(astrocalib_proc_vars['seg1xy_best'])}, {len(astrocalib_proc_vars['seg1xy_best'])/len(astrocalib_proc_vars['pos_seg'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['seg1xy_best'], pos2=astrocalib_proc_vars['seg2xy_best'], ax=axs[1,2])

            axs[0,3].set_title(f"d0 (1st)")
            get_pairs_d(astrocalib_proc_vars['pos_seg'], d_eps=astrocalib_proc_vars['d_eps'], bins=astrocalib_proc_vars['bins'], hist_range=astrocalib_proc_vars['hist_range'], ax=axs[0,3], doplot=True)
            axs[0,3].set_xlim([0,120])

            get_pairs_dxy(astrocalib_proc_vars['pos_seg'], d_eps=astrocalib_proc_vars['d_eps'], bins=astrocalib_proc_vars['bins'], hist_range=astrocalib_proc_vars['hist_range'], axs=[axs[1,3]], doplot=True)
            axs[1,3].set_title(f"disp (1st)")
            axs[1,3].set_xlim([0,120])

            fig.savefig(Path(redf.filedpropdir) / "astrometry_2_segmentation.png", bbox_inches="tight")
            fig.clf()
        else:
            fig = mplt.figure.Figure(figsize=(6,3), dpi=iop4conf.mplt_default_dpi)
            axs = fig.subplots(nrows=1, ncols=2)
            axs[0].set_title('Segmented Image')
            axs[0].imshow(astrocalib_proc_vars['segment_map'], cmap=astrocalib_proc_vars['segment_map'].cmap, origin="lower", interpolation="nearest")
            axs[1].set_title(f"Kron apertures (N={len(astrocalib_proc_vars['pos_seg'])})")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], ax=axs[1])
            astrocalib_proc_vars['seg_cat'].plot_kron_apertures(ax=axs[1], color='red', lw=1.5)

            fig.savefig(Path(redf.filedpropdir) / "astrometry_2_segmentation.png", bbox_inches="tight")
            fig.clf()

    # Summary image of DAOFind and pair finding results

    if 'pos_dao' in astrocalib_proc_vars.keys():

        logger.debug(f"{redf}: plotting astrometry summary image of daofind results")

        if astrocalib_proc_vars['has_pairs']:
            fig = mplt.figure.Figure(figsize=(12,6), dpi=iop4conf.mplt_default_dpi, layout="constrained")
            axs = fig.subplot_mosaic([["A", "B", "C", "D"], 
                                    ["A", "E", "F", "G"]])

            axs["A"].set_title(f"Sources (N={len(astrocalib_proc_vars['pos_dao'])})")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], pos1=astrocalib_proc_vars['pos_dao'], ax=axs["A"])

            axs["B"].set_title(f"Pairs (n={len(astrocalib_proc_vars['dao1'])}, {len(astrocalib_proc_vars['dao1'])/len(astrocalib_proc_vars['pos_dao'])*100:.1f}%)")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], pos1=astrocalib_proc_vars['dao1'], pos2=astrocalib_proc_vars['dao2'], ax=axs["B"])

            axs["E"].set_title(f"PairsXY (n={len(astrocalib_proc_vars['dao1xy'])}, {len(astrocalib_proc_vars['dao1xy'])/len(astrocalib_proc_vars['pos_dao'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['dao1xy'], pos2=astrocalib_proc_vars['dao2xy'], ax=axs["E"])

            axs["C"].set_title(f"BestPairs (n={len(astrocalib_proc_vars['dao1_best'])}, {len(astrocalib_proc_vars['dao1_best'])/len(astrocalib_proc_vars['pos_dao'])*100:.1f}%)")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], pos1=astrocalib_proc_vars['dao1_best'], pos2=astrocalib_proc_vars['dao2_best'], ax=axs["C"])

            axs["F"].set_title(f"BestPairsXY (n={len(astrocalib_proc_vars['dao1xy'])}, {len(astrocalib_proc_vars['dao1xy_best'])/len(astrocalib_proc_vars['pos_dao'])*100:.1f}%)")
            imshow_w_sources(redf.mdata, pos1=astrocalib_proc_vars['dao1xy_best'], pos2=astrocalib_proc_vars['dao2xy_best'], ax=axs["F"])

            axs['D'].set_title(f"d0 (1st)")
            get_pairs_d(astrocalib_proc_vars['pos_dao'], d_eps=astrocalib_proc_vars['d_eps'], bins=astrocalib_proc_vars['bins'], hist_range=astrocalib_proc_vars['hist_range'], ax=axs['D'], doplot=True)
            axs['D'].set_xlim([0,120])

            get_pairs_dxy(astrocalib_proc_vars['pos_dao'], d_eps=astrocalib_proc_vars['d_eps'], bins=astrocalib_proc_vars['bins'], hist_range=astrocalib_proc_vars['hist_range'], axs=[axs['G']], doplot=True)
            axs['G'].set_title(f"disp (1st)")
            axs['G'].set_xlim([0,120])

            fig.savefig(Path(redf.filedpropdir) / "astrometry_3_daofind.png", bbox_inches="tight")
            fig.clf()
        else:
            fig = mplt.figure.Figure(figsize=(3,3), dpi=iop4conf.mplt_default_dpi)
            ax = fig.subplots(nrows=1, ncols=1)
            ax.set_title(f"Sources (N={len(astrocalib_proc_vars['pos_dao'])})")
            imshow_w_sources(astrocalib_proc_vars['imgdata_bkg_substracted'], pos1=astrocalib_proc_vars['pos_dao'], ax=ax)
            fig.savefig(Path(redf.filedpropdir) / "astrometry_3_daofind.png", bbox_inches="tight")
            fig.clf()

    # Summary image of astrometry results

    if 'wcs1' in astrocalib_proc_vars.keys():
        logger.debug(f"{redf}: plotting astrometry summary image of astrometry results")

        fig = mplt.figure.Figure(figsize=(6,6), dpi=iop4conf.mplt_default_dpi)
        ax = fig.subplots(nrows=1, ncols=1, subplot_kw={'projection': astrocalib_proc_vars['wcs1']})
        plot_preview_astrometry(redf, **astrocalib_proc_vars, ax=ax, fig=fig, with_simbad=summary_kwargs['with_simbad'])
        fig.savefig(Path(redf.filedpropdir) / "astrometry_4_img_result.png", bbox_inches="tight")
        fig.clf()




def plot_finding_chart(target_src, fig=None, ax=None):

    from iop4lib.db import AstroSource
    from iop4lib.utils import get_simbad_sources

    radius = u.Quantity("6 arcmin")

    if fig is None:
        fig = plt.gcf()
    
    if ax is None:
        ax = plt.gca()

    simbad_sources = get_simbad_sources(target_src.coord, radius=radius, Nmax=5, exclude_self=False)
    calibrators = AstroSource.objects.filter(calibrates=target_src)
    
    for src in calibrators:
        ax.plot([src.coord.ra.deg], [src.coord.dec.deg], 'rx', alpha=1)
        ax.annotate(text=src.name, xy=(src.coord.ra.deg+0.001, src.coord.dec.deg), xytext=(+30,0), textcoords="offset pixels", color="red", fontsize=10, weight="bold", verticalalignment="center", horizontalalignment="left", arrowprops=dict(color="red", width=0.5, headwidth=1, headlength=3))
    
    for src in simbad_sources:
        ax.plot([src.coord.ra.deg], [src.coord.dec.deg], 'b+', alpha=1)
        ax.annotate(text=src.name, xy=(src.coord.ra.deg-0.001, src.coord.dec.deg), xytext=(-30,0), textcoords="offset pixels", color="blue", fontsize=10, weight="bold", verticalalignment="center", horizontalalignment="right", arrowprops=dict(color="blue", width=0.5, headwidth=1, headlength=3))

    ax.plot([target_src.coord.ra.deg], [target_src.coord.dec.deg], 'ro', alpha=1)

    # limits (center around target source)

    ax.set_xlim([target_src.coord.ra.deg - radius.to_value("deg"), target_src.coord.ra.deg + radius.to_value("deg")])
    ax.set_ylim([target_src.coord.dec.deg - radius.to_value("deg"), target_src.coord.dec.deg + radius.to_value("deg")])

    # labels

    ax.set_xlabel("RA [deg]")
    ax.set_ylabel("DEC [deg]")
    ax.set_title(f"{target_src.name} ({target_src.other_names})")

    # legend

    target_h = ax.plot([],[], 'ro', label=target_src.name)[0]
    simbad_h = ax.plot([],[], 'b+', label="SIMBAD sources")[0]
    calibrators_h = ax.plot([],[], 'rx', label="IOP4 Calibrators")[0]
    ax.legend(handles=[target_h, calibrators_h, simbad_h], loc="upper right")

    ax.grid(True, color='gray', ls='dashed')

    # secondary axes in hms and dms
    
    lims_ra = ax.get_xlim()
    ax_x2 = ax.twiny()
    ax_x2_ticks = ax.get_xticks()
    ax_x2.set_xticks(ax_x2_ticks)
    ax_x2.set_xticklabels([Angle(x, unit="deg").to_string(unit="hourangle", sep="hms") for x in ax_x2_ticks], fontsize=8)
    ax_x2.set_xlabel("RA [hms]")
    ax_x2.set_xlim(lims_ra)

    lims_dec = ax.get_ylim()
    ax_y2 = ax.twinx()
    ax_y2_ticks = ax.get_yticks()
    ax_y2.set_yticks(ax_y2_ticks)
    ax_y2.set_yticklabels([Angle(x, unit="deg").to_string(unit="deg", sep="dms") for x in ax_y2_ticks], fontsize=8)
    ax_y2.set_ylabel("DEC [dms]")
    ax_y2.set_ylim(lims_dec)


# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# django imports

# other imports
from abc import ABCMeta, abstractmethod

import os
import re
import numpy as np
import math
import astropy.io.fits as fits
import astropy.units as u
import itertools
import datetime
import glob
import astrometry
from photutils.centroids import centroid_sources, centroid_2dg

# iop4lib imports
from iop4lib.enums import *

# logging
import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from iop4lib.db import AstroSource, RawFit, ReducedFit
    from iop4lib.utils.astrometry import BuildWCSResult

class Instrument(metaclass=ABCMeta):
    """ Base class for instruments.
    
        Inherit this class to provide instrument specific functionality (e.g. classification of images,
        reduction, etc).

    """

    # Instrument identification (subclasses must implement these)

    @property
    @abstractmethod
    def name(self):
        """ The name of the instrument."""
        raise NotImplementedError

    @property
    @abstractmethod
    def telescope(self):
        """ The telescope this instrument is mounted on."""
        raise NotImplementedError

    @property
    @abstractmethod
    def instrument_kw(self):
        """ The keyword in the FITS header that identifies this instrument."""
        raise NotImplementedError

    # Instrument specific properties (subclasses must implement these)

    @property
    @abstractmethod
    def field_width_arcmin(self):
        """ Field width in arcmin."""
        raise NotImplementedError

    @property
    @abstractmethod
    def arcsec_per_pix(self):
        """ Pixel size in arcseconds per pixel."""
        raise NotImplementedError

    @property
    @abstractmethod
    def gain_e_adu(self):
        """ Gain in e-/ADU. Used to compute the error in aperture photometry."""
        raise NotImplementedError

    @property
    @abstractmethod
    def required_masters(self):
        r""" List of calibration frames needed.

            Cooled CCD cameras will only need `required_masters = ['masterbias', 'masterflat']` in the subclass, since dark current is close to zero. 
            If dark current is not negligible, set `required_masters = ['masterbias', 'masterdark', 'masterflat']` in the subclass.
        """
        raise NotImplementedError

    # Class methods (you should be using these from the Instrument class, not subclasses)

    @classmethod
    def get_known(cls):
        """ Return a list of all known instruments subclasses."""
        from .osn_cameras import RoperT90, AndorT90, AndorT150
        from .cafos import CAFOS
        from .dipol import DIPOL

        return [RoperT90, AndorT90, AndorT150, CAFOS, DIPOL]

    @classmethod
    def by_name(cls, name: str) -> 'Instrument':
        """
        Try to get instrument subclass by name, else raise Exception.
        """
        for instr in Instrument.get_known():
            if instr.name == name:
                return instr
        raise NotImplementedError(f"Instrument {name} not implemented.")
    
    # Common instrument functionality
    # You should be using these from the subclasses already
    # these don't need to be overriden in subclasses, but they can be

    # classification methods

    @classmethod
    def classify_rawfit(cls, rawfit):
        cls.check_instrument_kw(rawfit)
        cls.classify_juliandate_rawfit(rawfit)
        cls.classify_imgtype_rawfit(rawfit)
        cls.classify_band_rawfit(rawfit)
        cls.classify_obsmode_rawfit(rawfit)
        cls.classify_imgsize(rawfit)
        cls.classify_exptime(rawfit)
    
    @classmethod
    def check_instrument_kw(cls, rawfit):
        """ Check that the instrument keyword is correct. """
        if rawfit.header["INSTRUME"] not in cls.instrument_kw_L:
            raise ValueError(f"Raw fit file {rawfit.fileloc} has INSTRUME not in {cls.instrument_kw_L}.")

    @classmethod
    def classify_imgsize(cls, rawfit):
        """ Read the size of the image from the FITS header, and save it in rawfit.imgsize."""
        import astropy.io.fits as fits
        from iop4lib.db import RawFit

        with fits.open(rawfit.filepath) as hdul:
            if hdul[0].header["NAXIS"] == 2:
                sizeX = hdul[0].header["NAXIS1"]
                sizeY = hdul[0].header["NAXIS2"]
                rawfit.imgsize = f"{sizeX}x{sizeY}"
                return rawfit.imgsize
            else:
                raise ValueError(f"Raw fit file {rawfit.fileloc} has NAXIS != 2, cannot get imgsize.")
            
    @classmethod
    def classify_exptime(cls, rawfit):
        """
        EXPTIME is an standard FITS keyword, measured in seconds.
        """
        import astropy.io.fits as fits
        from iop4lib.db import RawFit

        with fits.open(rawfit.filepath) as hdul:
            rawfit.exptime = hdul[0].header["EXPTIME"]

    # reduction methods

    @classmethod
    def _get_header_hintobject_easy(self, rawfit) -> 'AstroSource' or None:
        r"""Get a hint for the AstroSource in this image from the header. 
        Return None if none found. 
        
        This method only tries to match the OBJECT keyword with the IAU name 
        format (`[0-9]*\+[0-9]*`) to the name of the sources in the DB.
        """
        
        from iop4lib.db import AstroSource

        object_header = rawfit.header["OBJECT"]
        
        matchs = re.findall(r".*?([0-9]*\+[0-9]*).*", object_header)
        if len(matchs) > 0:
            return AstroSource.objects.filter(name__contains=matchs[0]).first()
        else:
            return None
            
    @classmethod
    def _get_header_hintobject_hard(self, rawfit):
        """Get a hint for the AstroSource in this image from the header. 
        Return None if none found.

        This method tries to match the OBJECT keyword with each source in the
        DB, first trying all of them by their main names, then by their other 
        names.
        """

        from iop4lib.db import AstroSource

        catalog = AstroSource.objects.exclude(is_calibrator=True).all()

        pattern = re.compile(r"^([a-zA-Z0-9]{1,3}_[a-zA-Z0-9]+|[a-zA-Z0-9]{4,})(?=_|$)")
        
        obj_kw = rawfit.header['OBJECT']
        
        match = pattern.match(obj_kw)

        def get_invariable_str(s):
            return s.replace(' ', '').replace('-','').replace('+','').replace('_','').upper()

        if match:
            
            search_str = match.group(0)

            for source in catalog:
                if get_invariable_str(search_str) in get_invariable_str(source.name):
                    return source
                
            for source in catalog:
                if not source.other_names_list:
                    continue
                if any([get_invariable_str(search_str) in get_invariable_str(other_name) for other_name in source.other_names_list]):
                    return source
                
        return None
    
    @classmethod
    def get_header_hintobject(cls, rawfit):
        """Get the hint for the AstroSource in this image from the header.
        Return None if none found. 
        
        OBJECT is a standard keyword. This method will try to match the OBJECT 
        keyword, first using the more strict `_get_header_hintobject_easy`,
        and if that fails, using the more relaxed `_get_header_hintobject_hard`.
        """

        return cls._get_header_hintobject_easy(rawfit) or cls._get_header_hintobject_hard(rawfit)
    
    @classmethod
    @abstractmethod 
    def get_astrometry_position_hint(cls, rawfit, allsky=False, n_field_width=1.5, hintsep=None) -> astrometry.PositionHint:
        """ Get the position hint from the FITS header as an astrometry.PositionHint object. """        
        raise NotImplementedError

    @classmethod
    @abstractmethod 
    def get_astrometry_size_hint(cls, rawfit) -> astrometry.SizeHint:
        """ Get the size hint for this telescope / rawfit."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def has_pairs(cls, fit_instance: 'ReducedFit' or 'RawFit') -> bool:
        """ Indicates whether both ordinary and extraordinary sources are present 
        in the file. At the moment, this happens only for CAFOS polarimetry
        """
        raise NotImplementedError

    @classmethod
    def build_wcs(self, reducedfit: 'ReducedFit', shotgun_params_kwargs : dict = None, summary_kwargs : dict = None) -> 'BuildWCSResult':
        """ Build a WCS for a reduced fit from this instrument. 
        
        By default (Instrument class), this will just call the build_wcs_params_shotgun from iop4lib.utils.astrometry.

        Keyword Arguments
        -----------------
        shotgun_params_kwargs : dict, optional
            The parameters to pass to the shotgun_params function.
        summary_kwargs : dict, optional
            build_summary_images : bool, optional
                Whether to build summary images of the process. Default is True.
            with_simbad : bool, default True
                Whether to query and plot a few Simbad sources in the image. Might be useful to 
                check whether the found coordinates are correct. Default is True.
        """

        if shotgun_params_kwargs is None:
            shotgun_params_kwargs = dict()

        if summary_kwargs is None:
            summary_kwargs = {'build_summary_images':True, 'with_simbad':True}

        from iop4lib.utils.astrometry import build_wcs_params_shotgun
        from iop4lib.utils.plotting import build_astrometry_summary_images


        # if there is pointing mismatch information and it was not invoked with given allsky or position_hint,
        # fine-tune the position_hint or set allsky = True if appropriate.
        # otherwise leave it alone.

        if reducedfit.header_hintobject is not None and 'allsky' not in shotgun_params_kwargs and 'position_hint' not in shotgun_params_kwargs:

            pointing_mismatch = reducedfit.header_hintobject.coord.separation(reducedfit.header_hintcoord)

            # minimum of 1.5 the field width, and max the allsky_septhreshold

            hintsep = np.clip(1.1*pointing_mismatch, 
                                1.5*Instrument.by_name(reducedfit.instrument).field_width_arcmin*u.arcmin,
                                iop4conf.astrometry_allsky_septhreshold*u.arcmin)
            shotgun_params_kwargs["position_hint"] = [reducedfit.get_astrometry_position_hint(hintsep=hintsep)]

            # if the pointing mismatch is too large, set allsky = True if configured 

            if  pointing_mismatch.to_value('arcmin') > iop4conf.astrometry_allsky_septhreshold:
                if iop4conf.astrometry_allsky_allow:
                    logger.debug(f"{reducedfit}: large pointing mismatch detected ({pointing_mismatch.to_value('arcmin')} arcmin), setting allsky = True for the position hint.")
                    shotgun_params_kwargs["allsky"] = [True]
                    del shotgun_params_kwargs["position_hint"]


        build_wcs_result = build_wcs_params_shotgun(reducedfit, shotgun_params_kwargs, summary_kwargs=summary_kwargs)


        return build_wcs_result

    @classmethod
    def request_master(cls, rawfit, model, other_epochs=False):
        """ Searchs in the DB and returns an appropiate master bias / flat / dark for this rawfit. 
        
        Notes
        -----

        It takes into account the parameters (band, size, etc) defined in Master' margs_kwL; except 
        for exptime, since master calibration frames with different exptime can be applied.

        By default, it looks for masters in the same epoch, but if other_epochs is set to True, it
        will look for masters in other epochs. If more than one master is found, it returns the
        one from the closest night. It will print a warning even with other_epochs if it is more than 1
        week away from the rawfit epoch.
        
        If no master is found, it returns None.
        """

        from iop4lib.db import RawFit

        rf_vals = RawFit.objects.filter(id=rawfit.id).values().get()
        args = {k:rf_vals[k] for k in rf_vals if k in model.margs_kwL}
        
        args.pop("exptime", None) # exptime might be a building keywords (for flats and darks), but masters with different exptime can be applied
        args["epoch"] = rawfit.epoch # from .values() we only get epoch__id 

        master = model.objects.filter(**args, flags__hasnot=model.FLAGS.IGNORE).first()
        
        if master is None and other_epochs == True:
            args.pop("epoch")

            master_other_epochs = np.array(model.objects.filter(**args, flags__hasnot=model.FLAGS.IGNORE).all())

            if len(master_other_epochs) == 0:
                logger.debug(f"No {model._meta.verbose_name} for {args} in DB, None will be returned.")
                return None
            
            master_other_epochs_jyear = np.array([md.epoch.jyear for md in master_other_epochs])
            master = master_other_epochs[np.argsort(np.abs(master_other_epochs_jyear - rawfit.epoch.jyear))[0]]
            
            if (master.epoch.jyear - rawfit.epoch.jyear) > 7/365:
                logger.warning(f"{model._meta.verbose_name} from epoch {master.epoch} is more than 1 week away from epoch {rawfit.epoch}.")
                        
        return master


    @classmethod
    def associate_masters(cls, reducedfit, **masters_dict):
        """ Associate a masterbias, masterdark and masterflat to this reducedfit."""

        from iop4lib.db import ReducedFit, MasterBias, MasterDark, MasterFlat

        for (attrname, model) in zip(['masterbias', 'masterdark', 'masterflat'], [MasterBias, MasterDark, MasterFlat]):

            if attrname not in cls.required_masters:
                continue
            
            if masters_dict.get(attrname, None) is not None:
                setattr(reducedfit, attrname, masters_dict[attrname])
            else:
                if (master := reducedfit.rawfit.request_master(model)) is not None:
                    setattr(reducedfit, attrname, master)
                else:
                    logger.warning(f"{reducedfit}: {attrname} in this epoch could not be found, attemptying adjacent epochs.")
                    if (master := reducedfit.rawfit.request_master(model, other_epochs=True)) is not None:
                        setattr(reducedfit, attrname, master)
                    else:
                        logger.error(f"{reducedfit}: Could not find any {attrname}, not even in adjacent epochs.")
                        reducedfit.set_flag(ReducedFit.FLAGS.ERROR)

    @classmethod
    def apply_masters(cls, reducedfit):
        """ Apply the associated calibration frames to the raw fit to obtain the reduced fit."""
        
        import astropy.io.fits as fits

        logger.debug(f"{reducedfit}: applying masters")

        rf_data = fits.getdata(reducedfit.rawfit.filepath)
        mb_data = fits.getdata(reducedfit.masterbias.filepath)
        mf_data = fits.getdata(reducedfit.masterflat.filepath)

        if reducedfit.masterdark is not None:
            md_dark = fits.getdata(reducedfit.masterdark.filepath)
        else :
            logger.warning(f"{reducedfit}: no masterdark found, assuming dark current = 0, is this a CCD camera and it's cold?")
            md_dark = 0

        data_new = (rf_data - mb_data - md_dark*reducedfit.rawfit.exptime) / (mf_data)

        header_new = fits.Header()

        if not os.path.exists(os.path.dirname(reducedfit.filepath)):
            logger.debug(f"{reducedfit}: creating directory {os.path.dirname(reducedfit.filepath)}")
            os.makedirs(os.path.dirname(reducedfit.filepath))
        
        fits.writeto(reducedfit.filepath, data_new, header=header_new, overwrite=True)

    @classmethod
    def astrometric_calibration(cls, reducedfit: 'ReducedFit', **build_wcs_kwargs):
        """ Performs astrometric calibration on the reduced fit, giving it the appropriate WCS.

        If the are both ordinary and extraordinary sources in the field, one WCS will be built for each,
        and the will be saved in the first and second extensions of the FITS file.
        """

        # delete old astrometry info
        for fpath in glob.iglob(os.path.join(reducedfit.filedpropdir, "astrometry_*")):
            os.remove(fpath)

        # build the WCS
        build_wcs_result = cls.build_wcs(reducedfit, **build_wcs_kwargs)

        if build_wcs_result.success:

            logger.debug(f"{reducedfit}: saving WCSs to FITS header.")

            wcs1 = build_wcs_result.wcslist[0]

            header = fits.Header()

            header.update(wcs1.to_header(relax=True, key="A"))

            if reducedfit.has_pairs:
                wcs2 = build_wcs_result.wcslist[1]
                header.update(wcs2.to_header(relax=True, key="B"))

            # if available, save also some info about the astrometry solution
            if 'bm' in build_wcs_result.info:
                bm = build_wcs_result.info['bm']
                # adding HIERARCH avoids a warning, they can be accessed without HIERARCH
                header['HIERARCH AS_ARCSEC_PER_PIX'] = bm.scale_arcsec_per_pixel
                header['HIERARCH AS_CENTER_RA_DEG'] = bm.center_ra_deg
                header['HIERARCH AS_CENTER_DEC_DEG'] = bm.center_dec_deg

            # save the header to the file

            with fits.open(reducedfit.filepath, 'update') as hdul:
                hdul[0].header.update(header)

            # Save some extra info (not in the header)

            if not 'date' in build_wcs_result.info:
                build_wcs_result.info['date'] = datetime.datetime.now()

            try:
                if isinstance(reducedfit.astrometry_info, list):
                    reducedfit.astrometry_info = list(itertools.chain(reducedfit.astrometry_info, [build_wcs_result.info]))
                else:
                    reducedfit.astrometry_info = [build_wcs_result.info]
            except NameError:
                reducedfit.astrometry_info = [build_wcs_result.info]

        else:
            raise Exception(f"Could not perform astrometric calibration on {reducedfit}: {build_wcs_result=}")

    @classmethod
    def build_file(cls, reducedfit: 'ReducedFit', **build_wcs_kwargs):
        """ Builds the ReducedFit FITS file.

        Notes
        -----
        The file is built by:
        - applying master calibration frames.
        - astrometrically calibrate the reduced fit, giving it a WCS.
        - find the catalog sources in the field.
        """

        from iop4lib.db import AstroSource, ReducedFit

        logger.debug(f"{reducedfit}: building file")

        reducedfit.unset_flag(ReducedFit.FLAGS.BUILT_REDUCED)

        reducedfit.apply_masters()
        
        logger.debug(f"{reducedfit}: performing astrometric calibration")

        try:
            reducedfit.astrometric_calibration(**build_wcs_kwargs)
        except Exception as e:
            logger.error(f"{reducedfit}: could not perform astrometric calibration on {reducedfit}: {e}")
            reducedfit.set_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY)
            if reducedfit.auto_merge_to_db:
                reducedfit.save()
            raise e
        else:
            logger.debug(f"{reducedfit}: astrometric calibration was successful.")
            reducedfit.unset_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY)

            logger.debug(f"{reducedfit}: searching for sources in field...")
            sources_in_field = AstroSource.get_sources_in_field(fit=reducedfit)
            
            logger.debug(f"{reducedfit}: found {len(sources_in_field)} sources in field.")
            reducedfit.sources_in_field.set(sources_in_field, clear=True)
                
            reducedfit.set_flag(ReducedFit.FLAGS.BUILT_REDUCED)

        if reducedfit.auto_merge_to_db:
            reducedfit.save()

    @classmethod
    def compute_aperture_photometry(cls, redf, aperpix, r_in, r_out):
        """ Common aperture photometry method for all instruments."""

        from iop4lib.db.aperphotresult import AperPhotResult
        from iop4lib.utils.sourcedetection import get_bkg, get_segmentation
        from photutils.utils import circular_footprint
        from photutils.aperture import CircularAperture, CircularAnnulus, ApertureStats, aperture_photometry
        from photutils.utils import calc_total_error
        from astropy.stats import SigmaClip

        bkg_box_size = redf.mdata.shape[0]//10

        bkg = get_bkg(redf.mdata, filter_size=1, box_size=bkg_box_size)
        img = redf.mdata

        error = calc_total_error(img, bkg.background_rms, cls.gain_e_adu)

        apres_L = list()

        for astrosource in redf.sources_in_field.all():
            for pairs, wcs in (('O', redf.wcs1), ('E', redf.wcs2)) if redf.has_pairs else (('O',redf.wcs),):

                logger.debug(f"{redf}: computing aperture photometry for {astrosource}")

                wcs_px_pos = astrosource.coord.to_pixel(wcs)

                # check that the wcs px position is within (r_in+r_out)/2 from ther border of the image

                r_mid = (r_in + r_out) / 2

                if not (r_mid < wcs_px_pos[0] < img.shape[1] - r_mid and r_mid < wcs_px_pos[1] < img.shape[0] - r_mid):
                    logger.warning(f"{redf}: ({pairs}) image of {astrosource.name} is too close to the border, skipping aperture photometry.")
                    continue

                # correct position using centroid
                # choose a box size that is somewhat larger than the aperture
                # in case of pairs, cap box size that is somewhat smaller than the distance between pairs

                box_size = math.ceil(1.6 * aperpix)//2 * 2 + 1

                if redf.has_pairs:
                    _box_size_max = math.ceil(np.linalg.norm(Instrument.by_name(redf.instrument).disp_sign_mean))//2 * 2 - 1
                    box_size = min(box_size, _box_size_max)

                centroid_px_pos = centroid_sources(img, xpos=wcs_px_pos[0], ypos=wcs_px_pos[1], box_size=box_size, centroid_func=centroid_2dg)
                centroid_px_pos = (centroid_px_pos[0][0], centroid_px_pos[1][0])

                # check that the centroid position is within the borders of the image

                if not (r_mid < centroid_px_pos[0] < img.shape[1] - r_mid and r_mid < centroid_px_pos[1] < img.shape[0] - r_mid):
                    logger.warning(f"{redf}: centroid of the ({pairs}) image of {astrosource.name} is too close to the border, skipping aperture photometry.")
                    continue

                # log the difference between the WCS and the centroid
                wcs_diff = np.sqrt((centroid_px_pos[0] - wcs_px_pos[0])**2 + (centroid_px_pos[1] - wcs_px_pos[1])**2)
                
                logger.debug(f"ReducedFit {redf.id}: {astrosource.name} {pairs}: WCS centroid distance = {wcs_diff:.1f} px")

                ap = CircularAperture(centroid_px_pos, r=aperpix)
                annulus = CircularAnnulus(centroid_px_pos, r_in=r_in, r_out=r_out)

                annulus_stats = ApertureStats(redf.mdata, annulus, error=error, sigma_clip=SigmaClip(sigma=5.0, maxiters=10))
                ap_stats = ApertureStats(redf.mdata, ap, error=error)

                bkg_flux_counts = annulus_stats.median*ap_stats.sum_aper_area.value
                bkg_flux_counts_err = annulus_stats.sum_err / annulus_stats.sum_aper_area.value * ap_stats.sum_aper_area.value

                flux_counts = ap_stats.sum - annulus_stats.mean*ap_stats.sum_aper_area.value # TODO: check if i should use mean!
                flux_counts_err = ap_stats.sum_err

                apres = AperPhotResult.create(reducedfit=redf, 
                                      astrosource=astrosource, 
                                      aperpix=aperpix, 
                                      r_in=r_in, r_out=r_out,
                                      x_px=centroid_px_pos[0], y_px=centroid_px_pos[1],
                                      pairs=pairs, 
                                      bkg_flux_counts=bkg_flux_counts, bkg_flux_counts_err=bkg_flux_counts_err,
                                      flux_counts=flux_counts, flux_counts_err=flux_counts_err)
                
                apres_L.append(apres)

        return apres_L
    
    @classmethod
    def compute_relative_photometry(cls, redf: 'ReducedFit') -> None:
        """ Common relative photometry method for all instruments. """
        
        from iop4lib.db.aperphotresult import AperPhotResult
        from iop4lib.db.photopolresult import PhotoPolResult

        if redf.obsmode != OBSMODES.PHOTOMETRY:
            raise Exception(f"{redf}: this method is only for plain photometry images.")
        
        aperpix, r_in, r_out, fit_res_dict = cls.estimate_common_apertures([redf], reductionmethod=REDUCTIONMETHODS.RELPHOT)
        target_fwhm = fit_res_dict['mean_fwhm']
        
        if target_fwhm is None:
            logger.error("Could not estimate a target FWHM, aborting relative photometry.")
            return

        # 1. Compute all aperture photometries

        logger.debug(f"{redf}: computing aperture photometries for {redf} (target_fwhm = {target_fwhm:.1f} px, aperpix = {aperpix:.1f} px, r_in = {r_in:.1f} px, r_out = {r_out:.1f} px).")

        cls.compute_aperture_photometry(redf, aperpix, r_in, r_out)

        # 2. Compute relative polarimetry for each source (uses the computed aperture photometries)

        logger.debug(f"{redf}: computing relative photometry.")

        # 2. Compute the flux in counts and the instrumental magnitude
        
        photopolresult_L = list()
        
        for astrosource in redf.sources_in_field.all():

            qs_aperphotresult = AperPhotResult.objects.filter(reducedfit=redf, astrosource=astrosource, aperpix=aperpix, pairs="O")

            if not qs_aperphotresult.exists():
                logger.error(f"{redf}: no aperture photometry for source {astrosource.name} found, skipping relative photometry.")
                continue

            aperphotresult = qs_aperphotresult.first()

            result = PhotoPolResult.create(reducedfits=[redf], astrosource=astrosource, reduction=REDUCTIONMETHODS.RELPHOT)

            result.aperpix = aperpix
            result.bkg_flux_counts = aperphotresult.bkg_flux_counts
            result.bkg_flux_counts_err = aperphotresult.bkg_flux_counts_err
            result.flux_counts = aperphotresult.flux_counts
            result.flux_counts_err = aperphotresult.flux_counts_err

            # logger.debug(f"{self}: {result.flux_counts=}")

            if result.flux_counts is None: # when does this happen? when there is a source whose apertue falls partially outside the image? https://github.com/juanep97/iop4/issues/24
                logger.error(f"{redf}: during relative photometry, encountered flux_counts=None for source {astrosource.name}, aperphotresult {aperphotresult.id}!!!")
                result.flux_counts = np.nan
                result.flux_counts_err = np.nan

            if result.flux_counts <= 0.0:
                logger.warning(f"{redf}: negative flux counts encountered while relative photometry for {astrosource=} ??!! They will be nans, but maybe we should look into this...")

            result.mag_inst = -2.5 * np.log10(result.flux_counts) # np.nan if result.flux_counts <= 0.0
            result.mag_inst_err = math.fabs(2.5 / math.log(10) / result.flux_counts * result.flux_counts_err)

            # if the source is a calibrator, compute also the zero point
            if result.astrosource.is_calibrator:
                result.mag_known = getattr(result.astrosource, f"mag_{redf.band}")
                result.mag_known_err = getattr(result.astrosource, f"mag_{redf.band}_err", None) or 0.0

                if result.mag_known is None:
                    logger.warning(f"Relative Photometry over {redf}: calibrator {result.astrosource} has no magnitude for band {redf.band}.")
                    result.mag_zp = np.nan
                    result.mag_zp_err = np.nan
                else:
                    result.mag_zp = result.mag_known - result.mag_inst
                    result.mag_zp_err = math.sqrt(result.mag_inst_err**2 + result.mag_known_err**2)
            else:
                # if it is not a calibrator, we can not save the COMPUTED zp, it will be computed and the USED zp will be stored.
                result.mag_zp = None
                result.mag_zp_err = None

            result.aperphotresults.set([aperphotresult], clear=True)

            result.save()

            photopolresult_L.append(result)

        # 3. Compute the calibrated magnitudes

        for result in photopolresult_L:

            if result.astrosource.is_calibrator:
                continue

            # 3.a Average the zero points

            # get all the computed photopolresults that calibrate this source
            calibrator_results_L = [r for r in photopolresult_L if r.astrosource.calibrates.filter(pk=result.astrosource.pk).exists()]

            # Create an array with nan instead of None (this avoids the dtype becoming object)
            calib_mag_zp_array = np.array([result.mag_zp or np.nan for result in calibrator_results_L if result.astrosource.is_calibrator]) 
            calib_mag_zp_array = calib_mag_zp_array[~np.isnan(calib_mag_zp_array)]

            calib_mag_zp_array_err = np.array([result.mag_zp_err or np.nan for result in calibrator_results_L if result.astrosource.is_calibrator])
            calib_mag_zp_array_err = calib_mag_zp_array_err[~np.isnan(calib_mag_zp_array_err)]

            if len(calib_mag_zp_array) == 0:
                logger.error(f"{redf}: can not perform relative photometry on source {result.astrosource.name}, no calibrator zero-points found.")
                # [result.delete() for result in redf.photopolresults.all()]
                continue

            zp_avg = np.nanmean(calib_mag_zp_array)
            zp_std = np.nanstd(calib_mag_zp_array)

            zp_err = math.sqrt(np.sum(calib_mag_zp_array_err**2)) / len(calib_mag_zp_array_err)
            zp_err = math.sqrt(zp_std**2 + zp_err**2)

            # 3.b Compute the calibrated magnitude

            # save the zp (to be) used
            result.mag_zp = zp_avg
            result.mag_zp_err = zp_err

            # compute the calibrated magnitude
            result.mag = zp_avg + result.mag_inst
            result.mag_err = math.sqrt(result.mag_inst_err**2 + zp_err**2)

            result.save()
        
        # 5. Save the results

        for result in photopolresult_L:
            result.save()







    @classmethod
    def estimate_common_apertures(cls, reducedfits, reductionmethod=None, fit_boxsize=None, search_boxsize=(90,90), fwhm_min=2, fwhm_max=50):
        r"""estimate an appropriate common aperture for a list of reduced fits.
        
        It fits the target source profile in the fields and returns some multiples of the fwhm which are used as the aperture and as the inner and outer radius of the annulus for local bkg estimation).
        """

        from iop4lib.utils import fit_gaussian
        from iop4lib.db import AstroSource

        astrosource_S = set.union(*[set(redf.sources_in_field.all()) for redf in reducedfits])
        target_L = [astrosource for astrosource in astrosource_S if not astrosource.is_calibrator]

        logger.debug(f"{astrosource_S=}")

        if len(target_L) > 0:
            target = target_L[0]
        elif len(astrosource_S) > 0:
            target = astrosource_S.pop()
        else:
            return np.nan, np.nan, np.nan, {'mean_fwhm':np.nan, 'sigma':np.nan}
    
        fwhm_L = list()

        for redf in reducedfits:
            try:
                gaussian = fit_gaussian(px_start=redf.wcs.world_to_pixel(target.coord), redf=redf)
                fwhm = (2*np.sqrt(2*math.log(2))) * np.sqrt(gaussian[0].x_stddev.value**2+gaussian[0].y_stddev.value**2)
                if not (fwhm_min < fwhm < fwhm_max):
                    logger.warning(f"ReducedFit {redf.id} {target.name}: fwhm = {fwhm} px, skipping this reduced fit")
                    continue
                logger.debug(f"ReducedFit {redf.id} [{target.name}] Gaussian FWHM: {fwhm:.1f} px")
                fwhm_L.append(fwhm)
            except Exception as e:
                logger.warning(f"ReducedFit {redf.id} {target.name}: error in gaussian fit, skipping this reduced fit ({e})")
                continue

        if len(fwhm_L) > 0:
            mean_fwhm = np.mean(fwhm_L)
        else:
            logger.error(f"Could not find an appropriate aperture for Reduced Fits {[redf.id for redf in reducedfits]}, using standard fwhm of 3.5px")
            mean_fwhm = 3.5

        sigma = mean_fwhm / (2*np.sqrt(2*math.log(2)))
        r = sigma
        
        return 6.0*r, 7.0*r, 15.0*r, {'mean_fwhm':mean_fwhm, 'sigma':sigma}
    




from django.contrib import admin

# logging
import logging
logger = logging.getLogger(__name__)

class TextInputFilter(admin.SimpleListFilter):
    """
    Generic class for an input filter.
    """
    template = 'iop4admin/fits/textinput_filter.html'

    def lookups(self, request, model_admin):
        # Dummy, required to show the filter.
        return ((),)

    def choices(self, changelist):
        # Grab only the "all" option.
        all_choice = next(super().choices(changelist))
        all_choice['query_parts'] = (
            (k, v)
            for k, v in changelist.get_filters_params().items()
            if k != self.parameter_name
        )
        yield all_choice

class RawFitIdFilter(TextInputFilter):
    title = 'id'
    name = 'id'
    parameter_name = 'id'

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(id=self.value())
        
class RawFitNightFilter(TextInputFilter):
    title = 'Night'
    name = 'night'
    parameter_name = 'night'

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(epoch__night=self.value())
        
class RawFitFilenameFilter(TextInputFilter):
    title = 'Filename (regex)'
    name = 'filename'
    parameter_name = 'filename'

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(filename__iregex=self.value())

class RawFitTelescopeFilter(admin.SimpleListFilter):
    """
    Filter list for telescope in RawFit.
    """
    title = 'telescope'
    name = 'telescope'
    parameter_name = 'telescope'

    def lookups(self, request, model_admin):
        from iop4lib.telescopes import Telescope
        return ((t.name, t.name) for t in Telescope.get_known())

    def queryset(self, request, queryset):
        from iop4lib.telescopes import Telescope
        if (val := self.value()) is not None:
            return queryset.filter(epoch__telescope=Telescope.by_name(val).name)
        
class RawFitInstrumentFilter(admin.SimpleListFilter):
    """
    Filter list for instrument in RawFit.
    """
    title = 'instrument'
    name = 'instrument'
    parameter_name = 'instrument'

    def lookups(self, request, model_admin):
        from iop4lib.instruments import Instrument
        return ((i.name, i.name) for i in Instrument.get_known())
    
    def queryset(self, request, queryset):
        from iop4lib.instruments import Instrument
        if (val := self.value()) is not None:
            return queryset.filter(instrument=Instrument.by_name(val).name)

class RawFitFlagFilter(admin.SimpleListFilter):

    title = 'flags'
    name = 'flags'
    parameter_name = 'flags'

    def __init__(self, request, params, model, model_admin):
        super().__init__(request, params, model, model_admin)
        self.request = request
        self.model = model
        self.model_admin = model_admin

    def lookups(self, request, model_admin):
        from iop4lib.db import RawFit
        return ((flag.name, flag.value) for flag in RawFit.FLAGS)


    def queryset(self, request, queryset):
        if self.value() is None:
            return queryset
        
        selected_values = self.value().split(',')

        # Get the list of selected values
        selected_values = self.value().split(',')

        if "OPT_EXACT" in selected_values:
            OPT_EXACT = True
            selected_values.remove("OPT_EXACT")
        else:
            OPT_EXACT = False 


        if "OPT_NOTEXACT" in selected_values:
            OPT_NOTEXACT = True
            selected_values.remove("OPT_NOTEXACT")
        else:
            OPT_NOTEXACT = False

        if len(selected_values) == 0:
            return queryset

        # Filter objects using the custom method
        filtered_objects = []
        for obj in queryset:
            flags_yes= [self.model.FLAGS[val] for val in selected_values]
            flags_no = list(set(self.model.FLAGS)-set(flags_yes))

            flags_yes = [flag.value for flag in flags_yes]
            flags_no = [flag.value for flag in flags_no]

            if OPT_EXACT:
                if obj.has_flags(flags_yes) and obj.has_not_flags(flags_no):
                    filtered_objects.append(obj.id)
            elif OPT_NOTEXACT:
                if obj.has_flags(flags_yes):
                    filtered_objects.append(obj.id)

        return queryset.filter(id__in=filtered_objects)

    def choices(self, changelist):
        selected_values = self.value().split(',') if self.value() is not None else []

        # give default value
        if "OPT_EXACT" not in selected_values and "OPT_NOTEXACT" not in selected_values:
            selected_values.append("OPT_NOTEXACT")
        
        opt_exact_selected = "OPT_EXACT" in selected_values
        opt_notexact_selected = "OPT_NOTEXACT" in selected_values

        # AND OPTION
        new_values = selected_values.copy()

        if opt_exact_selected:
            new_values.remove("OPT_EXACT")
        else:
            new_values.append("OPT_EXACT")
            if opt_notexact_selected:
                new_values.remove("OPT_NOTEXACT")

        if len(new_values) == 0:
            query_string = changelist.get_query_string(remove=[self.parameter_name])
        else:
            query_string = changelist.get_query_string({self.parameter_name: ','.join(new_values)})

        yield {
            'selected': opt_exact_selected,
            'query_string': query_string,
            'display': 'exactly ' + ('(X)' if opt_exact_selected else '( )'),
        }

        # OR OPTION
        new_values = selected_values.copy()

        if opt_notexact_selected:
            new_values.remove("OPT_NOTEXACT")
        else:
            new_values.append("OPT_NOTEXACT")
            if opt_exact_selected:
                new_values.remove("OPT_EXACT")

        if len(new_values) == 0:
            query_string = changelist.get_query_string(remove=[self.parameter_name])
        else:
            query_string = changelist.get_query_string({self.parameter_name: ','.join(new_values)})
                
        yield {
            'selected': opt_notexact_selected,
            'query_string': query_string,
            'display': 'at least ' + ('(X)' if opt_notexact_selected else '( )'),
        }

        # VALUES 

        for lookup, title in self.lookup_choices:
            new_values = selected_values.copy()

            is_selected = lookup in selected_values

            if is_selected:
                new_values.remove(lookup)
            else:
                new_values.append(lookup)

            if len(new_values) == 0:
                query_string = changelist.get_query_string(remove=[self.parameter_name])
            else:
                query_string = changelist.get_query_string({self.parameter_name: ','.join(new_values)})
                
            yield {
                'selected': is_selected,
                'query_string':  query_string,
                'display': lookup, # title
            }




import sys, os, subprocess
GIT_COMMIT_HASH = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()
GIT_BRANCH = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()
GIT_DESCRIBE = subprocess.check_output(['git', 'describe', '--always'], cwd=os.path.dirname(os.path.realpath(__file__))).decode('ascii').strip()

# thumnail_gallery extension
from pathlib import Path
sys.path.insert(0, str(Path(__file__).parent / "sphinxext"))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# Get exact version from git

project = 'IOP4'
copyright = '2023, Juan Escudero Pedrosa'
author = 'Juan Escudero Pedrosa'
release = GIT_DESCRIBE

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'myst_nb',
    "nbsphinx",
    'thumbnail_gallery',
    'sphinxcontrib.bibtex',
    'sphinx.ext.autodoc',
    'sphinx.ext.linkcode',
    'numpydoc',
]

templates_path = ['_templates']
exclude_patterns = ['conf.py', '_build', 'sphinxext', '.*']
source_suffix = ['.rst', '.md']

nb_custom_formats = {
    '.py': ['jupytext.reads', {'fmt': 'py:percent'}],
}

nb_merge_streams = True # otherwise several prints stmts result in multiple chunks
nb_execution_timeout = 60 # seconds

# automatically extract thumbnails from notebooks (with sphinxext/thumbnail_gallery)

def get_thumbnails():
    from glob import glob
    from pathlib import Path

    nb_paths = glob(f"recipes/*.py")
    nb_names = [Path(nb_path).stem for nb_path in nb_paths]
    nbsphinx_thumbnails = {f"recipes/{nb_name}":f"_thumbnails/{nb_name}.png" for nb_name in nb_names}

    return nbsphinx_thumbnails

nbsphinx_thumbnails = get_thumbnails()

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# -- options for pydata theme

html_theme = "pydata_sphinx_theme"
pygments_style = 'sphinx'

html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/juanep97/iop4",
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        }
   ],
   "show_toc_level": 2,
   "show_nav_level": 2,
   "navigation_depth": 2,
}

html_static_path = ['_static']

html_css_files = [
    'pydata/custom.css',
]

# --- bibliography options
bibtex_bibfiles = ['citations.bib']
bibtex_reference_style = "author_year"

# -- doc options -------------------------------------------------

autodoc_default_options = {
    'member-order': 'groupwise',
    'no-undoc-members': True,
    'exclude-members': 'TimeoutException, DoesNotExist, MultipleObjectsReturned, get_deferred_fields, save_base, serializable_value, validate_unique, full_clean, clean_fields, prepare_database_save, unique_error_message, date_error_message, validate_constraints, refresh_from_db'
}

numpydoc_show_class_members = False 

# -- Configure IOP4 for imports from sphinx --
# IMPORTANT! Config must use the example file, or you will show your credentials in the docs!
# config_db=True as it needs to import the models.
import os, sys, pathlib
import iop4lib.config
from importlib import resources
iop4conf = iop4lib.Config(config_path=resources.files("iop4lib") / "config.example.yaml", config_db=True)
import iop4lib.db

# -- Add models' fields and their help_text to the documentation --

import inspect
from django.utils.html import strip_tags

def process_docstring(app, what, name, obj, options, lines):
    # This causes import errors if left outside the function
    from django.db import models

    # Only look at objects that inherit from Django's base model class
    if inspect.isclass(obj) and issubclass(obj, models.Model):
        # Grab the field list from the meta class
        fields = obj._meta.fields

        for field in fields:
            # Decode and strip any html out of the field's help text
            help_text = strip_tags(field.help_text)

            # Decode and capitalize the verbose name, for use if there isn't
            # any help text
            verbose_name = field.verbose_name

            if help_text:
                # Add the model field to the end of the docstring as a param
                # using the help text as the description
                lines.append(u':param %s: %s' % (field.attname, help_text))
            else:
                # Add the model field to the end of the docstring as a param
                # using the verbose name as the description
                lines.append(u':param %s: %s' % (field.attname, verbose_name))

            # Add the field's type to the docstring
            lines.append(u':type %s: %s' % (field.attname, type(field).__name__))

    # Return the extended docstring
    return lines

def setup(app):
    # Register the docstring processor with sphinx
    app.connect('autodoc-process-docstring', process_docstring)

def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object definition.

    This function is used by the sphinx.ext.linkcode extension.
    """
    if domain != 'py':
        return None

    # Get the module and object name
    module = sys.modules.get(info['module'])
    if module is None:
        return None
    obj = module.__dict__.get(info['fullname'])
    if obj is None:
        return None

    # Get the file location of the object's source code
    try:
        file_path = inspect.getsourcefile(obj)
    except Exception:
        return None

    # Convert the file path to a GitHub URL
    repo_url = 'https://github.com/juanep97/iop4'
    if file_path.startswith(os.path.commonprefix([iop4conf.basedir, file_path])):
        rel_path = os.path.relpath(file_path, iop4conf.basedir)
        url = f'{repo_url}/blob/{GIT_COMMIT_HASH}/{rel_path}'
        return url
    else:
        return None


import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

from django.db import models

import os
import datetime
import numpy as np
import astropy.io.fits as fits

from .fitfilemodel import FitFileModel
from ..enums import *
from .fields import FlagChoices, FlagBitField

import logging
logger = logging.getLogger(__name__)

class MasterDark(FitFileModel):
    """
    A class representing a master dark for an epoch.
    """

    margs_kwL = ['epoch', 'instrument', 'imgsize', 'exptime']

    imgtype = IMGTYPES.DARK

    # Database fields
    rawfits = models.ManyToManyField('RawFit', related_name='built_masterdarks')

    # flags

    class FLAGS(FlagChoices):
        IGNORE = 1 << 1
        
    flags = FlagBitField(choices=FLAGS.choices(), default=0)

    # fields corresponding to MasterDark kw arguments (margs_kwL)

    epoch = models.ForeignKey('Epoch', on_delete=models.CASCADE, related_name='masterdarks') 
    instrument = models.CharField(max_length=20, choices=INSTRUMENTS.choices)
    imgsize = models.CharField(max_length=12, null=True)
    exptime = models.FloatField(null=True)

    masterbias = models.ForeignKey('MasterBias', null=True, on_delete=models.CASCADE, related_name='masterdarks')

    class Meta:
        app_label = 'iop4api'
        verbose_name = "Master Dark"
        verbose_name_plural = "Master Darks"

    # Properties

    @property
    def filename(self):
        return f"masterdark_id{self.id}.fits"
    
    @property
    def filepath(self):
        return os.path.join(self.epoch.masterdarkdir, self.filename)
    
    @property
    def fileloc(self):
        return f"{self.epoch.epochname}/{self.filename}"

    # Filed Properties location
    
    @property
    def filedpropdir(self):
       return os.path.join(self.epoch.calibrationdir, 
                            self.filename +  '.d', 
                            self.__class__.__name__)
    
    # some helper properties

    @property
    def margs(self):
        """
        Return a dict of the arguments used to build this MasterDark.
        """
        return {'epoch':self.epoch, 'instrument':self.instrument, 'imgsize':self.imgsize, 'exptime':self.exptime}
    
    # repr and str

    @classmethod
    def margs2str(cls, margs):
       """
       Class method to build a nice string rep of the arguments of a MasterDark.
       """
       return f"{margs['epoch'].epochname} | {margs['instrument']} | {margs['imgsize']} | {margs['exptime']}s"

    def __repr__(self):
        return f"MasterDark.objects.get(id={self.id!r})"
    
    def __str__(self):
        return f"<MasterDark {self.id!r} | {MasterDark.margs2str(self.margs)}>"
    
    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self!r}")
        else:
            with p.group(4, f'<{self.__class__.__name__}(', ')>'):
                p.text(f"id: {self.id},")
                for k,v in self.margs.items():
                    p.breakable()
                    p.text(f"{k}: {v}")

    
    # Constructor
    @classmethod
    def create(cls, 
               rawfits=None, 
               masterbias=None,
               auto_merge_to_db=True,
               force_rebuild=True,
               **kwargs):
        """
            Create a MasterDark object for Epoch epoch.

            Parameters
            ----------
            epoch: Epoch
            instrument: str
            imgsize: "WidthxHeight" str
            band: str
            obsmode: str
            rotangle: float
            exptime: float

            Other Parameters
            ----------------
            rawfits: list of RawFit (optional)
            masterbias: MasterBias (optional)
            auto_merge_to_db: bool (optional) Default: True
        """

        #margs = {k:kwargs[k] for k in MasterDark.margs_kwL if k in kwargs}
        margs = {k:kwargs[k] for k in MasterDark.margs_kwL} # so it gives error if some margs kw missing

        from iop4lib.db import RawFit

        if (md := MasterDark.objects.filter(**margs).first()) is not None:
            logger.info(f"{md} exists, using it instead.")
        else:
            logger.info(f"A MasterDark entry for {MasterDark.margs2str(margs)} will be created.")
            md = cls(**margs)
            logger.debug("Saving MasterDark to DB so that it has an id.")
            md.save()

        if rawfits is None:
            rawfits = RawFit.objects.filter(imgtype=IMGTYPES.DARK, **margs)
            logger.debug(f"Found {len(rawfits)} dark raw files for {md}.")
        else:
            logger.debug(f"Using {len(rawfits)} dark raw files for {md}.")

        md.rawfits.set(rawfits)

        if masterbias is None:
            from .masterbias import MasterBias
            margs = {k:kwargs[k] for k in MasterBias.margs_kwL if k in kwargs}
            masterbias = MasterBias.objects.filter(**margs).first()    
            logger.debug(f"Using {masterbias} as MasterBias for {md}.")
            md.masterbias = masterbias        

        # Build the file
        if not md.fileexists or force_rebuild:
            logger.info(f"Building file")
            try:
                md.build_file()
            except Exception as e:
                logger.error(f"An error ocurred while building the MasterDark, deleting it and raising Exception.")
                md.delete()
                raise Exception(f"An error ocurred while building the MasterDark for {margs}: {e}")
            else:
                logger.info("MasterDark created successfully.")

        # merge to DB

        if auto_merge_to_db:
            md.save()

        return md

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_merge_to_db=True
    
    # Methods

    def build_file(self):
        logger.debug(f"Getting data from files")

        if self.rawfits.count() == 0:
            raise Exception(f"No rawfits for {self}")

        data_L = []
        for rf in self.rawfits.all():
            with fits.open(rf.filepath) as hdul:
                data = (hdul[0].data - self.masterbias.data) / self.exptime
                data_L.append(data)

        data = np.nanmedian(data_L, axis=0)

        logger.debug(f"Building header")

        header = fits.Header()

        header['TELESCOP'] = self.epoch.telescope
        header['NIGHT'] = self.epoch.yyyymmdd
        header['EPOCH'] = self.epoch.epochname
        header['IMGSIZE'] = self.imgsize
        header['IMGTYPE'] = 'masterdark'
        header['DATECREA'] = datetime.datetime.now(datetime.UTC).isoformat(timespec="milliseconds")
        header['NRAWFITS'] = self.rawfits.count()
        header['INSTRUME'] = self.instrument
        header['EXPTIME'] = self.exptime

        logger.debug(f"Building HDU")

        if not os.path.isdir(self.epoch.masterdarkdir):
            logger.debug(f"Creating directory {self.epoch.masterdarkdir}")
            os.makedirs(self.epoch.masterdarkdir)

        hdu = fits.PrimaryHDU(data, header=header)

        logger.debug(f"Writing MasterDark to {self.filepath}")
        
        hdu.writeto(self.filepath, overwrite=True)


import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)  

# other imports 

import os
import gc
from pathlib import Path
import numpy as np
import scipy as sp
import matplotlib as mplt
import matplotlib.pyplot as plt
import astrometry
import astropy.io.fits as fits
from astropy.wcs import WCS
from iop4lib.telescopes import Telescope
import astropy.units as u
from astropy.coordinates import SkyCoord
import multiprocessing
import functools
import itertools
import dataclasses

# iop4lib imports

from iop4lib.utils.sourcepairing import (get_pairs_d, get_pairs_dxy, get_best_pairs)
from iop4lib.utils.sourcedetection import (get_bkg, get_segmentation, get_cat_sources_from_segment_map)
from iop4lib.utils.plotting import build_astrometry_summary_images

# logging

import logging
logger = logging.getLogger(__name__)

import typing
if typing.TYPE_CHECKING:
    from iop4lib.db.reducedfit import ReducedFit

@dataclasses.dataclass
class BuildWCSResult():
    r"""
        'success': bool
            whether the appropiate WCSs was built successfully
        'wcslist': list 
            list of WCS objects built (usually one, two if there are extraordinary sources in the image)
        'info': dict or None 
            dict with extra information about the process

        Boolean evaluation of this object returns the value of 'success'.
    """
    success: bool
    wcslist: list[WCS] = dataclasses.field(default_factory=list)
    info: dict = dataclasses.field(default_factory=dict)   

    def __bool__(self):
        return self.success 


def build_wcs_params_shotgun(redf: 'ReducedFit', shotgun_params_kwargs : dict = None, hard : bool = False, summary_kwargs : dict = {'build_summary_images':True, 'with_simbad':True}) -> BuildWCSResult:
    """ Build the appropiate WCSs for a ReducedFit image, trying different parameters. See `build_wcs` for more info.

    Note: at the moment, this function tries source extraction with different combination of parameters and thresholds for 
    source extraction by calling a helper func (`_build_wcs_detect_and_try_solve`) with these parameters, which detects 
    the  sources with `photutils` image segmentation and tries to solve the WCS with the `astrometry.net` python wrapper. 
    The parameter combinations are chosen depending on the exposure time and the presence of pairs in the image.
    
    TODO:

    - Implement a more robust way to choose the parameters for source extraction such that the astrometry solver works  with less 
      attempts.
    - Explore other detectors and solvers if necessary to improve speed, sucess rate and accuracy.
    - Use pre-computed pair distances.
    """

    param_dicts_L = []

    params = dict()

    # Define the possible values of each parameter to `_build_wcs_detect_and_try_solve`

    # SENSIBLE VERSION

    # params["keep_n_seg"] = [150] # it is useless to try with much more than ~100 sources, if it is solvable it will be with the brightest sources, solver will take much much time otherwise
    # params["border_margin_px"] = [20] # sources very close to the border could be due to border effects
    # params["bkg_filter_size"] = [11, 7, 5, 3] # the higher the better, it will make the bkg much smoother (but a kernel too high will be very slow or get in computational errors, 11 is more than enough, 7 also produces pretty smooth results. Sometimes, smaller values (e.g. 5 or even 3 are better to capture very local variations). Bkg will be interpolated anyway according to bkg_box_size.
    # params["bkg_box_size"] = [8, 16, 32, 64] # 16 is low but produces nice backgrounds, anyway this is for source extraction, not for photometry, where higher sizes e.g. 32, 64, or even more would be more sensible. This is later interpolated, so with 16 you manage to capture more local variations of the background, important when there are sources of very different brightness in the image.
    # params["seg_fwhm"] = [1.0, 8.0] # the lower the better. Higher fwhm (8.0) will discard many sources due to noise, but will decrease precision. This is specially bad for images with pairs, as a pair might be detected as one source only in the middle.
    # params["seg_kernel_size"] = [None] # automatically set according to fwhm, default is 2*int(fwhmw)+1, which is good enough.
    # params["npixels"] = [64, 32, 16, 8] # higher will discard more fake sources, but will also discard real sources. However the larger, the centroid position might be more noisy. Better to keep it 8-64 depending on need (pairs vs no pairs, crowded vs no crowded, low exp vs high exp).
    # params["allsky"] = [False] # whether to look over all sky, sometimes the hint keywords are wrong (but vary rarely).
    # params["output_logodds_threshold"] = [80, 40, 21] # 21 is the default of the astrometry.net solver. Very rarely with very low exp. / very little sources, we might ned to put it to 14 to get a good match. LOWER IS NOT RECOMMENDED, as it might generate wrong matches.
    # params["n_rms_seg"] = [24.0, 12.0, 6.0, 3.0, 1.5, 1.0, 0.66] # might be worth it to set it depending on exposure time, very rarely even lower values might be helpful (0.5, 0.4).

    # FAST VERSION

    if redf.has_pairs:
        params["keep_n_seg"] = [300]
    else:
        params["keep_n_seg"] = [150]

    params["border_margin_px"] = [5]

    if redf.exptime <= 0.05:
        params["output_logodds_threshold"] = [14]
        params["n_rms_seg"] = [1.0, 0.8, 0.66]
    elif 0.05 < redf.exptime <= 0.1:
        params["output_logodds_threshold"] = [14]
        params["n_rms_seg"] = [1.2, 1.0, 0.8]
    elif 0.1 < redf.exptime <= 10.0:
        params["output_logodds_threshold"] = [14]
        params["n_rms_seg"] = [1.5, 1.2, 1.0]
    elif 10.0 < redf.exptime <= 30.0:
        params["output_logodds_threshold"] = [14]
        params["n_rms_seg"] = [3.0, 1.5, 1.2]
    else:
        params["output_logodds_threshold"] = [14]
        params["n_rms_seg"] = [6.0, 3.0, 1.5]

    params["bkg_filter_size"] = [11, 3] 
    params["bkg_box_size"] = [16, 8]
    params["seg_fwhm"] = [1.0]
    params["seg_kernel_size"] = [None]
    params["npixels"] = [32, 8, 16]
    params["allsky"] = [False]

    ## Substitute default params combinations with specified ones

    if shotgun_params_kwargs is not None:
        for k, v in shotgun_params_kwargs.items():
            params[k] = v

    ## Build a list of dictionaries with all the possible combinations of parameters

    param_dicts_L_light = []
    for values in itertools.product(*params.values()):
        param_dict = dict(zip(params.keys(), values))
        param_dicts_L_light.append(param_dict)

    ## filter out some senseless combinations

    param_dicts_L_light = list(filter(lambda x: x['bkg_filter_size'] < x['bkg_box_size'], param_dicts_L_light)) 

    ## add combinations to the list

    param_dicts_L = param_dicts_L + param_dicts_L_light 
    
    # HARD VERSION

    if hard:
        params["keep_n_seg"] = [300, 150]
        params["border_margin_px"] = [5, 20]
        params["bkg_filter_size"] = [11, 7, 5, 3]
        params["bkg_box_size"] = [64, 32, 16, 8]
        params["seg_fwhm"] = [1.0, 2.0, 4.0, 8.0]
        params["seg_kernel_size"] = [None]
        params["npixels"] = [64, 32, 16, 8]
        params["allsky"] = [False]
        params["output_logodds_threshold"] = [21, 14]
        params["n_rms_seg"] = [6.0, 3.0, 1.5, 1.0, 0.8, 0.66]

        if shotgun_params_kwargs is not None:
            for k, v in shotgun_params_kwargs.items():
                params[k] = v

        param_dicts_L_hard = []
        for values in itertools.product(*params.values()):
            param_dict = dict(zip(params.keys(), values))
            if param_dict not in param_dicts_L:
                param_dicts_L_hard.append(param_dict)

        param_dicts_L_hard = list(filter(lambda x: x['bkg_filter_size'] < x['bkg_box_size'], param_dicts_L_hard)) ## filter out some senseless combinations

        np.random.shuffle(param_dicts_L_hard) #inplace, so we try very different combinations first
        
        param_dicts_L = param_dicts_L + param_dicts_L_hard

    # Attempt to build the wcs with each combination of parameters

    logger.debug(f"{redf}: {len(param_dicts_L)} different combinations of parameters to try.")

    for i, params_dict in enumerate(param_dicts_L):
        logger.debug(f"{redf}: attempt {i+1} / {len(param_dicts_L)}, ({params_dict}) ...")

        build_wcs_result = _build_wcs_params_shotgun_helper(redf, **params_dict)
        # try: # if we want to allow exceptions, but if there's an error something is wrong, we should fix the cause of the underlying exception; different thing is that we could not solve astrometry
        #     build_wcs_result = _build_wcs_params_shotgun_helper(redf, **params_dict)
        # except Exception as e:
        #     logger.error(f"{redf}: some error ocurred during attempt {i+1} / {len(param_dicts_L)}, ({params_dict}), ignoring. Error: {e}")
        #     build_wcs_result = {'success': False}

        if build_wcs_result.success:
            logger.debug(f"{redf}: WCS built with attempt {i+1} / {len(param_dicts_L)} ({params_dict}).")
            break
    else:
        # if none worked
        logger.error(f"{redf}: could not solve astrometry with any of the {len(param_dicts_L)} default parameter combinations for source extraction.")
        return BuildWCSResult(success=False)

    # add the parameters that worked to the result
    build_wcs_result.info['params'] = params_dict

    # build summary images
    if summary_kwargs['build_summary_images']:
        logger.debug(f"{redf}: building summary images.")
        build_astrometry_summary_images(redf, build_wcs_result.info, summary_kwargs=summary_kwargs)

    # to remove unwanted info from the result
    to_save_from_info_kw_L = ['params', 'bm', 'seg_d0', 'seg_disp_sign', 'seg_disp_xy', 'seg_disp_sign_xy', 'seg_disp_xy_best']
    build_wcs_result.info = {k:build_wcs_result.info[k] for k in to_save_from_info_kw_L if k in build_wcs_result.info}
    build_wcs_result.info['logodds'] = build_wcs_result.info['bm'].logodds
    build_wcs_result.info['method'] = 'shotgun'

    return build_wcs_result




def _build_wcs_params_shotgun_helper(redf, has_pairs=None,
        bkg_filter_size = 11,
        bkg_box_size = 16,
        seg_kernel_size = None,
        npixels = 32,
        seg_fwhm = 1.0,
        n_rms_seg = 1.0,
        keep_n_seg = 200,
        border_margin_px = 20,
        dx_eps=None,
        dy_eps=None,
        d_eps=None,
        dx_min=None,
        dx_max=None,
        dy_min=None,
        dy_max=None,
        d_min=None,
        d_max=None,
        bins=None,
        hist_range=None,
        position_hint=None, size_hint=None, allsky=False,
        output_logodds_threshold=21) -> BuildWCSResult:
    """ helper func, see build_wcs_params_shotgun for more info. """

    imgdata = redf.mdata

    if has_pairs is None:
        has_pairs = redf.has_pairs

    if size_hint is None:
        size_hint = redf.get_astrometry_size_hint()

    if position_hint is None:
        position_hint = redf.get_astrometry_position_hint(allsky=allsky)

    if has_pairs:
        if bins is None:
            bins = int( 0.75 * max(imgdata.shape) )
        if hist_range is None:
            hist_range = (0, min(imgdata.shape))

    # Background substraction

    bkg = get_bkg(imgdata, filter_size=bkg_filter_size, box_size=bkg_box_size)
    imgdata_bkg_substracted = imgdata - bkg.background

    # Image Segmentation

    seg_threshold = n_rms_seg * bkg.background_rms
    segment_map, convolved_data = get_segmentation(imgdata_bkg_substracted, fwhm=seg_fwhm, kernel_size=seg_kernel_size, npixels=npixels, threshold=seg_threshold)

    if segment_map is None:
        logger.debug(f"{redf}: No segments found, returning early.")
        return BuildWCSResult(success=False)
    
    seg_cat, pos_seg, tb = get_cat_sources_from_segment_map(segment_map, imgdata_bkg_substracted, convolved_data)
    
    logger.debug(f"{redf}: {len(pos_seg)=}")

    if keep_n_seg is not None and len(pos_seg) > keep_n_seg:
        logger.debug(f"{redf}: Keeping only {keep_n_seg} brightest segments.")
        pos_seg = pos_seg[:keep_n_seg]

    if border_margin_px is not None:
        logger.debug(f"{redf}: Removing segments within {border_margin_px} px from border.")
        pos_seg = [pos for pos in pos_seg if ( (border_margin_px < pos[0] < imgdata.shape[1]-border_margin_px) and (border_margin_px < pos[1] < imgdata.shape[0]-border_margin_px))]

    # Pair finding with results from image segmentation
    
    if has_pairs:
        seg1, seg2, seg_d0, seg_disp_sign = get_pairs_d(pos_seg, bins=bins, hist_range=hist_range, d_min=d_min, d_eps=d_eps, d_max=d_max)
        logger.debug(f"{redf}: seg pairs -> {len(seg1)} ({len(seg1)/len(pos_seg)*100:.1f}%), seg_disp_sign={seg_disp_sign}")
        seg1_best, seg2_best, seg_disp_best, seg_disp_sign_best = get_best_pairs(seg1, seg2, seg_disp_sign)
        logger.debug(f"{redf}: seg pairs best -> {len(seg1_best)} ({len(seg1_best)/len(pos_seg)*100:.1f}%), seg_disp_sign_best={seg_disp_sign_best}")
        seg1xy, seg2xy, seg_disp_xy, seg_disp_sign_xy = get_pairs_dxy(pos_seg, bins=bins, hist_range=hist_range, dx_min=dx_min, dx_max=dx_max, dy_min=dy_min, dy_max=dy_max, dx_eps=dx_eps, dy_eps=dy_eps)
        logger.debug(f"{redf}: seg pairs xy -> {len(seg1xy)}, disp_sign_xy={seg_disp_sign_xy}")
        seg1xy_best, seg2xy_best, seg_disp_xy_best, seg_disp_sign_xy_best = get_best_pairs(seg1xy, seg2xy, seg_disp_sign_xy)
        logger.debug(f"{redf}: seg pairs xy best -> {len(seg1xy_best)} ({len(seg1xy_best)/len(pos_seg)*100:.1f}%), seg_disp_sign_xy_best={seg_disp_sign_xy_best}")

    # Solve astrometry 

    bm = None

    if has_pairs: ## Attempt both with D pairs and XY pairs
        attempts = ((f"Seg Best XY Pairs (n={len(seg1xy_best)})", seg1xy_best, seg_disp_sign_xy_best),
                    (f"Seg Best D Pairs (n={len(seg1_best)})", seg1_best, seg_disp_sign_best),)
    else: ## Use the positions of the segments
        attempts = ((f"Seg Pos (n={len(pos_seg)})", pos_seg, None),)
    
    for msg, stars, disp_sign in attempts:
        if len(stars) == 0:
            logger.debug(f"{redf}: can not solve astrometry with {msg} because no pairs where found.")
            continue

        logger.debug(f"{redf}: trying to solve astrometry with {msg} ({output_logodds_threshold=}).")

        solution = solve_astrometry(stars, size_hint=size_hint, position_hint=position_hint, output_logodds_threshold=output_logodds_threshold)

        if solution.has_match():
            bm = solution.best_match()
            break
        else:
            logger.debug(f"{redf}: could not solve astrometry with {msg}.")

    ## If not, desist and return early; else we continue and build and save the wcs.

    if bm is None:
        return BuildWCSResult(success=False)
    else:
        logger.debug(f"{redf}: {msg} worked.")
        logger.debug(f"{redf}: {bm.index_path=}")
        logger.debug(f"{redf}: {bm.center_ra_deg=}")
        logger.debug(f"{redf}: {bm.center_dec_deg=}")
        logger.debug(f"{redf}: {bm.scale_arcsec_per_pixel=}")
        logger.debug(f"{redf}: {bm.logodds=}")
    
    # Build WCS from the best match

    wcs1 = WCS(bm.wcs_fields)

    if has_pairs:
        # Build WCS for pairs (just displace the center pixel by the disp_sign)
        wcs2 = wcs1.deepcopy()
        wcs2.wcs.crpix[0] += disp_sign[0]
        wcs2.wcs.crpix[1] += disp_sign[1]

    # save results and return

    return BuildWCSResult(success=True, 
                          wcslist=[wcs1, wcs2] if has_pairs else [wcs1], 
                          info=_save_astrocalib_proc_vars(locals()))




def _save_astrocalib_proc_vars(locals_dict):
    """ helper func to save all variables related to the source detection to a single dict. """

    astrocalib_proc_vars = dict()

    save_list = [
        'msg',
        'has_pairs',
        'bkg_box_size', 'bkg_filter_size',
        'bkg',
        'imgdata_bkg_substracted',
        'convolved_data',
        'seg_threshold', 'pos_seg', 'segment_map', 'seg_cat',
        'solution', 'bm', 'wcs1', 'header1',
        'stars', 'disp_sign',
    ]

    if locals_dict['has_pairs']:
        save_list += [
        'msg',
        'wcs2',
        'hist_range', 'bins', 'd_eps',
        'seg1', 'seg2', 'seg_d0', 'seg_disp_sign',
        'seg1_best', 'seg2_best', 'seg_disp_best', 'seg_disp_sign_best',
        'seg1xy', 'seg2xy', 'seg_disp_xy', 'seg_disp_sign_xy',
        'seg1xy_best', 'seg2xy_best', 'seg_disp_xy_best', 'seg_disp_sign_xy_best',
        ]

    astrocalib_proc_vars.update({k: v for k, v in locals_dict.items() if k in save_list})

    return astrocalib_proc_vars












# TODO: astrometry package .solve() leaks memory, which does not get freed even after gc.collect()
# and we ran many trials, until the leak builds up and crashesh the computer. As a workaround, this
# wrappers wraps the function solve_astrometry in such a way that it runs in a separate process,
# therefore all memory leaked is freed after the function ends. The wrapper allows us to use the function
# as if it was not wrapped, and it is transparent to the user. A better solution would be to fix the
# memory leak in the astrometry package.
#
# The wrapper starts the processes and waits for it to finish, and then returns the result.
#
#"""
# The wrapper uses the `multiprocess` package instead of `multiprocessing` because the latter does 
# not allow for daemon processes to have children, and solve_astrometry is invoked from a Pool of 
# processes when iop4conf.nthreads > 1, and Pool's processes are daemons. Another 
# alternative would have been to start the concurrent threads using Process instead of Pool 
# with daemon=False. 
#
# Using PoolExecutor in both ocassions (for concurrency and for leak containment) does allow
# to have two nested PoolExecutors (see below) but then the wrapper can not be a decorator 
# because multiprocessing can not pickle the function (dill can, which is why mutltiprocess
# works).

def _queue_wrapper(q, f, *args, **kwargs):
    q.put(f(*args, **kwargs))

def run_in_process(f):
    from functools import wraps
    import multiprocess
    @wraps(f)
    def wrapper(*args, **kwargs):
        ctx = multiprocess.context.ForkContext()
        q = ctx.Queue()
        p = ctx.Process(target=_queue_wrapper, args=(q, f) + args, kwargs=kwargs)
        p.start()
        # then get the result
        result = q.get()
        # then join
        p.join()
        return result

    return wrapper
#"""

"""
def solve_astrometry(*args, **kwargs):
    from functools import wraps
    import concurrent.futures, multiprocessing
    
    with concurrent.futures.ProcessPoolExecutor(mp_context=multiprocessing.get_context('fork')) as executor:
        future = executor.submit(_solve_astrometry, *args, **kwargs)
        result = future.result()
    return result

# invokes the astrometry solver with given stars positions and size and position hints
# def _solve_astrometry ... # if using the two nested pool excutors
"""

@run_in_process
def solve_astrometry(stars, size_hint=None, position_hint=None, output_logodds_threshold=21, sip=False, positional_noise_pixels=1):
    """ Actually run Astrometry solver with the given stars and position and size hints. """

    def logodds_callback(logodds_list: list[float]) -> astrometry.Action:
        """ Return early if there is a match with logodds greater than 90. """
        if np.amax(logodds_list) > 90: return astrometry.Action.STOP
        return astrometry.Action.CONTINUE
    
    solver = astrometry.Solver(
        astrometry.series_5200.index_files(
            cache_directory=iop4conf.astrometry_cache_path,
            scales={0,1,2,3,4}, ## 1,4 might be faster
        ) #+ astrometry.series_4200.index_files(
          #  cache_directory=iop4conf.astrometry_cache_path,
          #  scales={0,1,2,3,4},
        #)
    )

    if not sip: # default astrometry.net solver
        solution_parameters = astrometry.SolutionParameters(logodds_callback=logodds_callback, 
                                                            output_logodds_threshold=output_logodds_threshold, 
                                                            sip_order=0, tune_up_logodds_threshold=None,
                                                            positional_noise_pixels=positional_noise_pixels)
    else: # but for a telescope we do not expect sip to be important
        solution_parameters = astrometry.SolutionParameters(logodds_callback=logodds_callback, output_logodds_threshold=output_logodds_threshold)

    # TODO: this code was here to confirm that it was .solve() that was leaking memory.
    # It was confirmed it, and I am leaving it here for future debugging.
    #from iop4lib.utils import get_mem_parent_from_child, get_total_mem_from_child
    #child_mem_before_solve = get_mem_parent_from_child()
    #total_mem_before_solve = get_total_mem_from_child()

    solution = solver.solve(
        stars=stars,
        size_hint=size_hint,
        position_hint=position_hint,
        solution_parameters=solution_parameters,
    )

    #child_mem_after_solve = get_mem_parent_from_child()
    #total_mem_after_solve = get_total_mem_from_child()
    #logger.warning(f"Child memory usage before solve from pid {os.getpid()}: {child_mem_before_solve/1024**2:.0f} MB, after solve {child_mem_after_solve/1024**2:.0f} MB, difference {(child_mem_after_solve-child_mem_before_solve)/1024**2:.0f} MB")
    #logger.warning(f"Total memory usage before solve from pid {os.getpid()}: {total_mem_before_solve/1024**2:.0f} MB, after solve {total_mem_after_solve/1024**2:.0f} MB, difference {(total_mem_after_solve-total_mem_before_solve)/1024**2:.0f} MB")

    return solution


""" Configuration for pytest tests.

Django pytest tests require giving the real database path. Also, we would like 
to use the configured astrometry index files path, if it was already configured 
in the system were tests are being run, to avoid downloading them again 
unnecessarily.

The configuration involves reading the real IOP4 config, changing the datadir 
and database paths to the test ones, and creating a test config file. It will 
also check whether the right version of the test dataset is available, otherwise
it will download it. Test submodules must be configured to use the created test
config file, e.g. with
```
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)
```

The test datadir is removed after the tests are run.
"""

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

# other imports
import os
import sys
import pytest
import yaml
import hashlib
from pathlib import Path

TESTDATA_MD5SUM = '8b065bbcd32a3e3c10620f0d49c5c782'
TESTDATA_FPATH = str(Path(f"~/iop4testdata.{TESTDATA_MD5SUM}.tar.gz").expanduser())
TEST_CONFIG = str(Path("~/iop4testdata/config.tests.yaml").expanduser())
TEST_DATADIR = str(Path("~/iop4testdata").expanduser())
TEST_DB_PATH = str(Path("~/iop4testdata/test.db").expanduser())

def pytest_configure():

    from django.conf import settings

    settings.configure(
        INSTALLED_APPS=[
            'iop4api',
            # required to test the web interface, so the DB has user table
            "iop4admin",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
        DATABASES = {
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": iop4conf.db_path,
                "TEST": {"NAME":TEST_DB_PATH},
            }
        },
        DEBUG = False,
        SECRET_KEY = "fake-test-key",
    )
    
    iop4conf.configure(db_path=TEST_DB_PATH, datadir=TEST_DATADIR)
    iop4conf.config_path = TEST_CONFIG
    iop4conf.datadir = TEST_DATADIR
    iop4conf.db_path = TEST_DB_PATH
    
@pytest.fixture(scope="session")
def testdata(request):
    setUpClass()
    request.addfinalizer(tearDownClass)
    
def setUpClass():

    # check if test data is available; otherwise download it
    if not os.path.exists(TESTDATA_FPATH):
        print("Downloading test dataset")
        import requests
        with requests.get(f"https://vhega.iaa.es/iop4/testdata/{Path(TESTDATA_FPATH).name}", stream=True) as r:
            r.raise_for_status()
            with open(TESTDATA_FPATH, 'wb') as f:
                for chunk in r.iter_content(chunk_size=None):
                    f.write(chunk)        

    # prepare test data dir
    try:
        os.makedirs(Path(iop4conf.datadir).parent, exist_ok=True)
    except Exception as e:
        raise Exception(f"Error creating test data directory: {e}")
    
    # unpack test data
    if os.system(f"tar -xzf {TESTDATA_FPATH} -C {Path(TEST_DATADIR).parent}") != 0:
        raise Exception("Error unpacking test dataset")
    
    # create test config file
    try:
        with open(TEST_CONFIG, 'w') as f:
            f.write(yaml.safe_dump(dict(iop4conf)))
    except Exception as e:
        raise Exception(f"Error creating test config file: {e}")

    # check if test data dir is the right one, else exist inmediately (to avoid deleting wrong dir)
    if Path(iop4conf.datadir).name != "iop4testdata":
        print("DANGER!")
        sys.exit(-1)

def tearDownClass():
    # check if test data dir is the right one, else exist inmediately (to avoid deleting wrong dir)
    if Path(iop4conf.datadir).name != "iop4testdata":
        print("DANGER!")
        sys.exit(-1)
    # remove test data dir
    os.system(f"rm -rf {iop4conf.datadir}")



import pytest
from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import re
from pathlib import Path
from django.test import client
from django.test import override_settings
from django.contrib import auth

# logging
import logging
logger = logging.getLogger(__name__)

# we need to make iop4site module available as iop4site, not as iop4site.iop4site
# we also need to load all the rest of settings that we usually don't use
import sys
sys.path.insert(0, str(Path(__file__).parents[1] / "iop4site/"))
import iop4site.settings
settings_dict = {key:value for key,value in vars(iop4site.settings).items() if re.match(r'^[A-Z][A-Z_0-9]*$', key)}
# We want to test the production scenario
settings_dict["DEBUG"] = False
settings_dict["ALLOWED_HOSTS"] = ['testserver']

@override_settings(**settings_dict)
def test_index(client):
    """Test the index page"""
    response = client.get('/')
    assert response.status_code == 200
    assert b"Juan Escudero Pedrosa" in response.content

@override_settings(**settings_dict)
@pytest.mark.django_db(transaction=True)
def test_login(client):
    """Test the login and login page"""

    response = client.get('/iop4/login/')
    assert response.status_code == 200
    assert b"Username" in response.content
    assert b"Password" in response.content

    # create a test user
    from django.contrib.auth.models import User
    user = User.objects.create_user(username='testuser', password='testpassword')
    user.save()

    # try to login
    response = client.post('/iop4/api/login/', 
                           {'username': 'testuser', 'password': 'testpassword'},
                            follow=True)
    
    # check that the user is authenticated
    user = auth.get_user(client)
    assert user.is_authenticated

    # check that the explore tab is present
    client.get('/iop4/explore/')
    assert response.status_code == 200
    assert b"Explore data" in response.content

@override_settings(**settings_dict)
@pytest.mark.django_db(transaction=True)
def test_failed_login(client):
    """Test that the login fails with wrong credentials"""

    response = client.get('/iop4/login/')
    assert response.status_code == 200
    assert b"Username" in response.content
    assert b"Password" in response.content

    # create a test user
    from django.contrib.auth.models import User
    user = User.objects.create_user(username='testuser', password='testpassword')
    user.save()

    # try to login
    response = client.post('/iop4/api/login/', 
                           {'username': 'testuser', 'password': 'wrongtestpassword'},
                            follow=True)
    
    # check that the user is not authenticated
    user = auth.get_user(client)
    assert not user.is_authenticated

    # check that the explore tab is not present
    client.get('/iop4/explore/')
    assert response.status_code == 200
    assert not b"Explore data" in response.content